from datetime import datetime
from enum import Enum
import sys
import time
from typing import Any, Dict, List, Optional, Tuple, Union

from packaging import version

from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.build import Build
from anyscale.client.openapi_client.models.session import Session
from anyscale.cluster_compute import (
    get_cluster_compute_from_name,
    get_selected_cloud_id_or_default,
)
from anyscale.conf import MINIMUM_RAY_VERSION
from anyscale.links import DOCS_CLUSTER
from anyscale.sdk.anyscale_client import (
    ComputeTemplateConfig,
    CreateCluster,
    CreateComputeTemplate,
    StartClusterOptions,
    UpdateCluster,
)
from anyscale.sdk.anyscale_client.models.compute_template import ComputeTemplate
from anyscale.sdk.anyscale_client.models.session_state import SessionState
from anyscale.util import generate_slug, get_endpoint, wait_for_session_start
from anyscale.utils.connect_helpers import get_cluster, list_entities
from anyscale.utils.name_utils import gen_valid_name


# Max number of auto created clusters.
MAX_CLUSTERS = 40

# The type of the dict that can be passed to create a cluster env.
# e.g., {"base_image": "anyscale/ray-ml:1.1.0-gpu"}
CLUSTER_ENV_DICT_TYPE = Dict[str, Union[str, List[str]]]

# The cluster compute type. It can either be a string, eg my_template or a dict,
# eg, {"cloud_id": "id-123" ...}
CLUSTER_COMPUTE_DICT_TYPE = Dict[str, Any]

# Commands used to build Ray from source. Note that intermediate stages will
# be cached by the app config builder.
BUILD_STEPS = [
    "git clone https://github.com/ray-project/ray.git",
    "curl -fsSL https://bazel.build/bazel-release.pub.gpg | gpg --dearmor > bazel.gpg",
    "sudo mv bazel.gpg /etc/apt/trusted.gpg.d/",
    'echo "deb [arch=amd64] https://storage.googleapis.com/bazel-apt stable jdk1.8" | sudo tee /etc/apt/sources.list.d/bazel.list',
    "sudo apt-get update && sudo apt-get install -y bazel=3.2.0",
    'cd ray/python && sudo env "PATH=$PATH" python setup.py develop',
    "pip uninstall -y ray",
]

# Default minutes for autosuspend.
DEFAULT_AUTOSUSPEND_TIMEOUT = 120

# Default docker images to use for connect clusters.
def _get_base_image(image: str, ray_version: str, cpu_or_gpu: str) -> str:
    py_version = "".join(str(x) for x in sys.version_info[0:2])
    if py_version not in ["36", "37", "38"]:
        raise ValueError(f"No default docker image for py{py_version}")
    return f"anyscale/{image}:{ray_version}-py{py_version}-{cpu_or_gpu}"


class PrepareClusterAction(str, Enum):
    START = "START"
    CREATE = "CREATE"
    UPDATE = "UPDATE"
    WAIT = "WAIT"
    NO_OP = "NO_OP"


class PrepareClusterBlock:
    def __init__(  # noqa: PLR0913
        self,
        project_id: str,
        cluster_name: Optional[str],
        autosuspend_timeout: Optional[int],
        allow_public_internet_traffic: Optional[bool],
        needs_update: bool,
        cluster_compute_name: Optional[str],
        cluster_compute_dict: Optional[CLUSTER_COMPUTE_DICT_TYPE],
        cloud_name: Optional[str],
        build_pr: Optional[int],
        force_rebuild: bool,
        build_commit: Optional[str],
        cluster_env_name: Optional[str],
        cluster_env_dict: Optional[CLUSTER_ENV_DICT_TYPE],
        cluster_env_revision: Optional[int],
        ray: Any,
        log_output: bool = True,
    ):
        """
        Class to prepare a cluster (create and start if required) so it can be
        connected to via ray.client. Relevant information about the cluster is
        stored in the instance variables.
        """
        self.project_id = project_id
        self.cluster_name = cluster_name
        self.autosuspend_timeout = autosuspend_timeout
        self.allow_public_internet_traffic = allow_public_internet_traffic
        self.needs_update = needs_update
        self.cluster_compute_name = cluster_compute_name
        self.cluster_compute_dict = cluster_compute_dict
        self.cloud_name = cloud_name
        self.build_pr = build_pr
        self.force_rebuild = force_rebuild
        self.build_commit = build_commit
        self.cluster_env_name = cluster_env_name
        self.cluster_env_dict = cluster_env_dict
        self.cluster_env_revision = cluster_env_revision
        self._ray = ray

        auth_api_client = get_auth_api_client(log_output=log_output)
        self.api_client = auth_api_client.api_client
        self.anyscale_api_client = auth_api_client.anyscale_api_client

        self.log = BlockLogger(log_output=log_output)
        self.block_label = "PrepareCluster"
        self.log.open_block(self.block_label, block_title="Preparing the cluster")

    def prepare(self):
        prepare_cluster_action = self._get_prepare_cluster_action(
            self.project_id, self.cluster_name, self.needs_update
        )

        if prepare_cluster_action in {
            PrepareClusterAction.CREATE,
            PrepareClusterAction.START,
            PrepareClusterAction.UPDATE,
        }:
            existing_terminated_cluster = None
            if self.cluster_name:
                existing_terminated_cluster = get_cluster(
                    self.anyscale_api_client, self.project_id, self.cluster_name
                )

            self.cluster_env_name = self._build_cluster_env_if_needed(
                self.project_id,
                self.build_pr,
                self.build_commit,
                self.cluster_env_dict,
                self.cluster_env_name,
                self.force_rebuild,
            )

            # If the cluster build is not explicitly provided by the user
            # then we use the existing cluster's build id (if the cluster
            # already exists). Otherwise, we get the default build ID.
            if self.cluster_env_name or existing_terminated_cluster is None:
                build_id = self._get_cluster_build(
                    self.cluster_env_name, self.cluster_env_revision
                ).id
            else:
                build_id = existing_terminated_cluster.build_id

            cloud_id = get_selected_cloud_id_or_default(
                api_client=self.api_client,
                anyscale_api_client=self.anyscale_api_client,
                cloud_name=self.cloud_name,
            )
            if (
                (existing_terminated_cluster is not None)
                and (existing_terminated_cluster.cloud_id is not None)
                and (self.cloud_name is not None)  # user explicitly passed cloud var.
                and (
                    cloud_id != existing_terminated_cluster.cloud_id
                )  # cloud_id is the id of self.cloud_name
            ):
                raise ValueError(
                    "Current cluster already has a cloud. Changing it to a new cloud is not allowed. If you need a new cloud, please create a new cluster."
                )

            # If the user explicitly passes a cluster compute (dict or name),
            # Or if the cluster does not exist,
            # Or if the cluster already exist but the user is changing clouds,
            # Then get the updated cluster compute.
            # Otherwise, get the compute template of the existing terminated cluster.
            if (
                self.cluster_compute_name
                or self.cluster_compute_dict
                or existing_terminated_cluster is None
                or (
                    self.cloud_name is not None  # user explicitly passed cloud var.
                    and cloud_id
                    != existing_terminated_cluster.cloud_id  # cloud_id is the id of self._cloud_name
                )
            ):
                compute_template_id = self._get_cluster_compute_id(
                    self.project_id,
                    self.cluster_compute_name,
                    self.cluster_compute_dict,
                    self.cloud_name,
                )
            else:
                compute_template_id = existing_terminated_cluster.compute_template_id

            self._wait_for_app_build(self.project_id, build_id)

            self.cluster_name = self._start_or_create_cluster(
                project_id=self.project_id,
                build_id=build_id,
                compute_template_id=compute_template_id,
                cluster_name=self.cluster_name,
                autosuspend_timeout=self.autosuspend_timeout,
                allow_public_internet_traffic=self.allow_public_internet_traffic,
            )
        else:
            assert (
                self.cluster_name
            ), f"{self.cluster_name} is None, yet _derive_prepare_cluster_action returned {prepare_cluster_action}."
            cluster = get_cluster(
                self.anyscale_api_client, self.project_id, self.cluster_name
            )
            assert (
                cluster
            ), f"Cluster {self.cluster_name} not found, yet _derive_prepare_cluster_action returned {prepare_cluster_action}."
            self._validate_new_cluster_compute_and_env_match_existing_cluster(
                project_id=self.project_id, running_cluster=cluster,
            )
            if prepare_cluster_action == PrepareClusterAction.WAIT:
                cluster = get_cluster(
                    self.anyscale_api_client, self.project_id, self.cluster_name
                )
                wait_for_session_start(
                    self.project_id,
                    cluster.name,
                    self.api_client,
                    log=self.log,
                    block_label=self.block_label,
                )

            self.log.info(
                f"Cluster {BlockLogger.highlight(self.cluster_name)} is currently running.",
                block_label=self.block_label,
            )
            self.log.info("Connecting to this cluster:", block_label=self.block_label)
            url = get_endpoint(f"/projects/{self.project_id}/clusters/{cluster.id}")
            self._log_cluster_configs(
                cluster, cluster.build_id, cluster.compute_template_id, url,
            )

        self.log.close_block(self.block_label)

    def _start_or_create_cluster(
        self,
        project_id: str,
        build_id: str,
        compute_template_id: str,
        cluster_name: Optional[str],
        autosuspend_timeout: Optional[int],
        allow_public_internet_traffic: Optional[bool],
    ) -> str:
        """Create/Start a cluster based on its current state and passed args.

        Args:
            project_id (str): The project to use.
            build_id (str): Build to start cluster with.
            compute_template_id (str): Compute template to start cluster with
            cluster_name (Optional[str]): If specified, the given cluster
                will be created or updated as needed. Otherwise the cluster
                name will be picked automatically.
            autosuspend_timeout (Optional[int]): Autosuspend value of cluster
            allow_public_internet_traffic (bool): Whether to allow public internet
                traffic to serve endpoints
        Returns:
            The name of the cluster to connect to.
        """
        ray_cli = self._ray.util.client.ray
        # If a cluster name is not specified, user wants to start a new cluster.
        did_user_not_specify_cluster_name = not cluster_name
        if did_user_not_specify_cluster_name:
            # Try to generate an auto-incrementing cluster name from the first 100 clusters.
            # If there are more than 40 clusters with, we will generate a random name.
            # If the name is already taken, we will re-try with a different random name.
            results = list_entities(
                self.anyscale_api_client.list_sessions, project_id, max=100
            )
            self.log.debug("-> Starting a new cluster")
            used_names = [s.name for s in results]
            for i in range(MAX_CLUSTERS):
                name = f"cluster-{i}"
                if name not in used_names:
                    cluster_name = name
                    self.log.debug("Starting cluster", cluster_name)
                    break

        self.log.debug(
            f"Updating {cluster_name} to use build id {build_id} and compute template id {compute_template_id}"
        )
        # TODO(ekl): race condition here since "up" breaks the lock.
        if ray_cli.is_connected():
            self._ray.util.disconnect()
        # Update cluster.
        self.log.debug("Starting cluster with sdk and compute config.")

        cluster, start_required = self._create_or_update_session_data(
            cluster_name,
            project_id,
            build_id,
            compute_template_id,
            autosuspend_timeout,
            bool(allow_public_internet_traffic),
            did_user_not_specify_cluster_name,
        )
        return self._start_cluster_if_required(
            cluster,
            start_required,
            project_id,
            build_id,
            compute_template_id,
            allow_public_internet_traffic,
        )

    def _start_cluster_if_required(
        self,
        cluster: Session,
        start_required: bool,
        project_id: str,
        build_id: str,
        compute_template_id: str,
        allow_public_internet_traffic: Optional[bool],
    ) -> str:
        """Start the cluster if required.

        Args:
            cluster (Session): The cluster to start.
            start_required (bool): Whether the cluster.
            project_id (str): The project to use.
            build_id (str): Build to start cluster with.
            compute_template_id (str): Compute template to start cluster with
            allow_public_internet_traffic (bool): Whether to allow public internet
                traffic to serve endpoints
        """
        url = get_endpoint(f"/projects/{project_id}/clusters/{cluster.id}")

        if start_required:
            self.log.debug(
                "Note that restarting the cluster will not change the configurations correctly "
                "in AIOA or GCP."
            )

        self._log_cluster_configs(cluster, build_id, compute_template_id, url)

        if start_required:
            self.anyscale_api_client.start_cluster(
                cluster.id,
                StartClusterOptions(
                    cluster_environment_build_id=build_id,
                    cluster_compute_id=compute_template_id,
                    allow_public_internet_traffic=allow_public_internet_traffic,
                ),
            )

            wait_for_session_start(
                project_id,
                cluster.name,
                self.api_client,
                log=self.log,
                block_label=self.block_label,
            )
            self.log.debug(f"Cluster {cluster.name} finished starting. View at {url}")

        return str(cluster.name)

    def _log_cluster_configs(
        self, cluster: Session, build_id: str, compute_template_id: str, url: str,
    ) -> None:
        """Prints information about the cluster."""

        cluster_env = self.anyscale_api_client.get_build(build_id).result
        # The SDK uses application_template_id but this is really just the cluster_env_name.
        cluster_env_name = (
            cluster_env.application_template_id + ":" + str(cluster_env.revision)
        )

        compute_config: ComputeTemplate = self.anyscale_api_client.get_compute_template(
            compute_template_id
        ).result
        compute_config_name = compute_config.name
        compute_config_config = compute_config.config

        left_pad = " " * 2
        self.log.info(
            f"{left_pad}{'cluster id:': <30}{cluster.id}", block_label=self.block_label
        )
        self.log.info(
            f"{left_pad}{'cluster environment:': <30}{cluster_env_name}",
            block_label=self.block_label,
        )
        self.log.info(
            f"{left_pad}{'cluster environment id:': <30}{build_id}",
            block_label=self.block_label,
        )
        self.log.info(
            f"{left_pad}{'cluster compute:': <30}{compute_config_name}",
            block_label=self.block_label,
        )
        self.log.info(
            f"{left_pad}{'cluster compute id:': <30}{compute_template_id}",
            block_label=self.block_label,
        )

        def get_minutes_output(field: Optional[int]) -> str:
            if field and field > 0:
                return f"{field} minutes"
            return "disabled"

        self.log.info(
            f"{left_pad}{'idle termination:': <30}{get_minutes_output(cluster.idle_timeout)}",
            block_label=self.block_label,
        )
        self.log.info(
            f"{left_pad}{'maximum uptime:': <30}{get_minutes_output(compute_config_config.maximum_uptime_minutes)}",
            block_label=self.block_label,
        )

        self.log.info(f"{left_pad}{'link:': <30}{url}", block_label=self.block_label)

    def _create_or_update_session_data(  # noqa: PLR0913
        self,
        cluster_name: Optional[str],
        project_id: str,
        build_id: str,
        compute_template_id: str,
        idle_timeout: Optional[int],
        allow_public_internet_traffic: bool = False,
        did_user_not_specify_cluster_name: bool = False,
    ) -> Tuple[Session, bool]:
        """
        If cluster name is None, creates a new cluster with a generates a cluster name.

        Creates new cluster with build_id and compute_template_id if cluster
        with `cluster_name` doesn't already exist. Otherwise, update the
        `idle_timeout` of the existing cluster if provided.
        """

        start_required = True

        if not cluster_name:
            # Generate a cluster name
            cluster_name = f"cluster-{generate_slug()}"

        cluster = get_cluster(self.anyscale_api_client, project_id, cluster_name)

        if did_user_not_specify_cluster_name:
            while cluster:
                # User wants to create a new cluster but cluster with name already exists,
                # generate a new name until we have one that's not taken.
                cluster_name = f"cluster-{generate_slug()}"
                cluster = get_cluster(
                    self.anyscale_api_client, project_id, cluster_name
                )

        if not cluster:
            # Create a new cluster if there is no existing cluster with the given cluster_name
            # This will occur if the user did not specify a cluster name, or if a cluster with
            # the user specified cluster_name does not exist.
            self.log.info(
                f"Cluster {BlockLogger.highlight(cluster_name)} does not exist. A new cluster will be created.",
                block_label=self.block_label,
            )
            self.log.info(
                f"Starting cluster {BlockLogger.highlight(cluster_name)}:",
                block_label=self.block_label,
            )
            # User didn't specify idle timeout, use the default autosuspend timeout
            if idle_timeout is None:
                idle_timeout = DEFAULT_AUTOSUSPEND_TIMEOUT
            create_cluster_data = CreateCluster(
                name=cluster_name,
                project_id=project_id,
                cluster_environment_build_id=build_id,
                cluster_compute_id=compute_template_id,
                idle_timeout_minutes=idle_timeout,
                allow_public_internet_traffic=allow_public_internet_traffic,
            )
            self.anyscale_api_client.create_cluster(create_cluster_data)
            cluster = get_cluster(self.anyscale_api_client, project_id, cluster_name)
            assert cluster is not None
        else:
            # Get the existing cluster and update the idle_timeout if required
            if cluster.state == "Running":
                start_required = self._validate_new_cluster_compute_and_env_match_existing_cluster(
                    project_id, cluster, print_warnings=False
                )
                if start_required:
                    self.log.info(
                        f"Cluster {BlockLogger.highlight(cluster_name)} is currently running, "
                        "but the cluster configurations specified in `ray.init()` indicate an "
                        f"update should occur. This will restart cluster {BlockLogger.highlight(cluster_name)}.",
                        block_label=self.block_label,
                    )
                    self.log.info(
                        f"Restarting cluster {BlockLogger.highlight(cluster_name)}:",
                        block_label=self.block_label,
                    )
                else:
                    self.log.info(
                        f"Cluster {BlockLogger.highlight(cluster_name)} is currently running.",
                        block_label=self.block_label,
                    )
                    self.log.info(
                        "Connecting to this cluster:", block_label=self.block_label
                    )
            else:
                self.log.info(
                    f"Cluster {BlockLogger.highlight(cluster_name)} exists but not running. This cluster will be restarted.",
                    block_label=self.block_label,
                )
                self.log.info(
                    f"Restarting cluster {BlockLogger.highlight(cluster_name)}:",
                    block_label=self.block_label,
                )
            if idle_timeout:
                self.anyscale_api_client.update_cluster(
                    cluster.id, UpdateCluster(idle_timeout_minutes=idle_timeout)
                )

        return cluster, start_required

    def _get_prepare_cluster_action(
        self, project_id: str, cluster_name: Optional[str], needs_update: bool
    ) -> PrepareClusterAction:
        """
        Returns the next cluster action based on the cluster state.
        """
        if not cluster_name:
            return PrepareClusterAction.CREATE
        cluster = get_cluster(self.anyscale_api_client, project_id, cluster_name)
        if not cluster:
            # Unconditionally create the cluster if it isn't up.
            return PrepareClusterAction.CREATE
        elif cluster.state == SessionState.STARTINGUP:
            return PrepareClusterAction.WAIT
        elif cluster.state != SessionState.RUNNING:
            # Unconditionally start the cluster if it isn't up.
            return PrepareClusterAction.START
        elif needs_update:
            return PrepareClusterAction.UPDATE
        else:
            return PrepareClusterAction.NO_OP

    def _get_default_cluster_env_build(self) -> Build:
        """
        Get the default cluster env build based on the local python and ray versions.
        """
        py_version = "".join(str(x) for x in sys.version_info[0:2])
        if sys.version_info.major == 3 and sys.version_info.minor == 10:
            py_version = "310"
        if py_version not in ["36", "37", "38", "39", "310"]:
            raise ValueError(
                "No default cluster env for py{}. Please use a version of python between 3.6 and 3.8.".format(
                    py_version
                )
            )
        ray_version = self._ray.__version__
        if version.parse(ray_version) < version.parse(MINIMUM_RAY_VERSION):
            raise ValueError(
                f"No default cluster env for Ray version {ray_version}. Please upgrade "
                f"to a version >= {MINIMUM_RAY_VERSION}."
            )
        if "dev0" in ray_version:
            raise ValueError(
                f"Your locally installed Ray version is {ray_version}. "
                "There is no default cluster environments for nightly versions of Ray."
            )
        try:
            build = self.api_client.get_default_cluster_env_build_api_v2_builds_default_py_version_ray_version_get(
                f"py{py_version}", ray_version
            ).result
            return build
        except Exception:  # noqa: BLE001
            raise RuntimeError(
                f"Failed to get default cluster env for Ray: {ray_version} on Python: py{py_version}"
            )

    def _get_cluster_build(
        self, cluster_env_name: Optional[str], cluster_env_revision: Optional[int],
    ) -> Build:
        """Returns the build of the cluster to be created.
        By default we return the default cluster env, unless the user overrides.
        """
        if cluster_env_name:
            build = self._get_cluster_env_build(cluster_env_name, cluster_env_revision)
        else:
            try:
                build = self._get_default_cluster_env_build()
            except ValueError as e:
                # A default cluster env was not found
                raise ValueError(
                    "Because you did not specify a cluster environment, we "
                    "attempted to use a default cluster environment but could "
                    "not find one that matches your requirements. "
                    "To specify a cluster environment, use the cluster_env "
                    "argument to ray.init() or the ANYSCALE_CLUSTER_ENV environment variable. "
                    "If one does not already exist, you can create a new cluster environment "
                    f"with your dependencies by following these instructions: {DOCS_CLUSTER}."
                ) from e

        return build

    def _get_cluster_env_build(
        self, cluster_env_name: str, clust_env_revision: Optional[int]
    ) -> Build:
        """
        Get the build for the specified cluster environment. If the cluster env revsion is
        not provided, use the latest revision.
        """
        app_template_id = None
        cluster_environments = list_entities(
            self.anyscale_api_client.list_app_configs,
            filters={"name_contains": cluster_env_name},
        )
        for cluster_env in cluster_environments:
            if cluster_env.name == cluster_env_name:
                app_template_id = cluster_env.id
        if not app_template_id:
            raise RuntimeError(
                f"Cluster Environment '{cluster_env_name}' not found. See environments at "
                f'{get_endpoint("/configurations/?tab=cluster-env")}.'
            )
        builds = list_entities(self.anyscale_api_client.list_builds, app_template_id)

        build_to_use = None
        if clust_env_revision:
            for build in builds:
                if build.revision == clust_env_revision:
                    build_to_use = build

            if not build_to_use:
                raise RuntimeError(
                    "Revision {} of cluster environment '{}' not found.".format(
                        clust_env_revision, cluster_env_name
                    )
                )
        else:
            latest_build_revision = -1
            for build in builds:
                if build.revision > latest_build_revision:
                    latest_build_revision = build.revision
                    build_to_use = build
            self.log.debug(
                "Using latest revision {} of {}".format(
                    latest_build_revision, cluster_env_name
                )
            )
        assert build_to_use  # for mypy
        return build_to_use

    def _build_cluster_env_if_needed(
        self,
        project_id: str,
        build_pr: Optional[int],
        build_commit: Optional[str],
        cluster_env_dict: Optional[CLUSTER_ENV_DICT_TYPE],
        cluster_env_name: Optional[str],
        force_rebuild: bool,
    ):
        """Builds a new cluster env on the fly if a cluster env dict is provided by the user
        or if the user wants to build ray from source."""
        if build_pr or build_commit:
            cluster_env_name = self._build_app_config_from_source(
                project_id, build_pr, build_commit, force_rebuild
            )

        elif cluster_env_dict:
            # Replacing ":" with "-" because the cluster env name cannot include ":"
            cluster_env_name = (
                cluster_env_name.replace(":", "-")
                if cluster_env_name
                else "anonymous_cluster_env-{}".format(
                    datetime.now().strftime("%Y-%m-%d_%H-%M-%S-%f")
                )
            )
            self.log.info(
                f"Building a new docker image for cluster environment {cluster_env_name} with "
                "the provided cluster environment dictionary.",
                block_label=self.block_label,
            )
            self.anyscale_api_client.create_app_config(
                {
                    "name": cluster_env_name,
                    "project_id": project_id,
                    "config_json": cluster_env_dict,
                }
            )
        return cluster_env_name

    def _build_app_config_from_source(
        self,
        project_id: str,
        build_pr: Optional[int],
        build_commit: Optional[str],
        force_rebuild: bool,
    ) -> str:
        """
        Build a cluster environment from a build PR and/or commit.
        """
        config_name = "ray-build"
        if build_pr:
            config_name += f"-{build_pr}"
        if build_commit:
            config_name += f"-{build_commit}"
        # Force creation of a unique app config.
        if force_rebuild:
            config_name += f"-{int(time.time())}"
        app_templates = list_entities(
            self.anyscale_api_client.list_app_configs,
            filters={"name_contains": config_name},
        )
        found = any(a.name == config_name for a in app_templates)
        if not found:
            build_steps = BUILD_STEPS.copy()
            # Add a unique command to bust the Makisu cache if needed.
            # Otherwise we could end up caching a previous fetch.
            build_steps.append(f"echo UNIQUE_ID={config_name}")
            if build_pr:
                build_steps.append(
                    "cd ray && git fetch origin pull/{}/head:target && "
                    "git checkout target".format(build_pr)
                )
            if build_commit:
                build_steps.append(f"cd ray && git checkout {build_commit}")
            build_steps.append(
                'cd ray/python && sudo env "PATH=$PATH" python setup.py develop'
            )
            self.anyscale_api_client.create_app_config(
                {
                    "name": config_name,
                    "project_id": project_id,
                    "config_json": {
                        "base_image": _get_base_image("ray", "nightly", "cpu"),
                        "debian_packages": ["curl", "unzip", "zip", "gnupg"],
                        "post_build_cmds": build_steps,
                    },
                }
            )
        return config_name

    def _wait_for_app_build(self, project_id: str, build_id: str) -> Build:
        """
        Block until cluster env finishes building.
        """
        has_logged = False
        while True:
            build = self.anyscale_api_client.get_build(build_id).result
            if build.status in ["pending", "in_progress"]:
                if not has_logged:
                    url = get_endpoint(
                        f"projects/{project_id}/app-config-details/{build_id}"
                    )
                    self.log.info(
                        f"Waiting for cluster env to be built (see {url} for progress)...",
                        block_label=self.block_label,
                    )
                    has_logged = True
                time.sleep(10.0)
            elif build.status in ["failed", "pending_cancellation", "canceled"]:
                raise RuntimeError(
                    "Cluster env status is '{}', please select another revision".format(
                        build.status
                    )
                )
            else:
                assert build.status == "succeeded"
                return build

    def _get_cluster_compute_id(
        self,
        project_id: str,
        cluster_compute_name: Optional[str],
        cluster_compute_dict: Optional[CLUSTER_COMPUTE_DICT_TYPE],
        cloud_name: Optional[str],
    ) -> str:
        """Returns the compute template ids of the cluster to be created.
        By default we return the default cluster compute, unless the user overrides.
        """
        # get cluster compute template
        if cluster_compute_name:
            compute_template_id = get_cluster_compute_from_name(
                cluster_compute_name, self.api_client
            ).id
        else:
            # If the user specifies _cluster_compute_dict use it, otherwise
            # use the default cluster compute template.
            if cluster_compute_dict:
                cluster_compute_class = ComputeTemplateConfig(**cluster_compute_dict)
                config_object = cluster_compute_class
            else:
                cloud_id = get_selected_cloud_id_or_default(
                    api_client=self.api_client,
                    anyscale_api_client=self.anyscale_api_client,
                    cloud_name=cloud_name,
                )
                config_object = self.anyscale_api_client.get_default_compute_config(
                    cloud_id
                ).result
            compute_template_id = self._register_compute_template(
                project_id, config_object
            )
        return compute_template_id

    def _register_compute_template(
        self, project_id: str, config_object: ComputeTemplateConfig
    ) -> str:
        """
        Register compute template with a default name and return the compute template id."""
        created_template = self.api_client.create_compute_template_api_v2_compute_templates_post(
            create_compute_template=CreateComputeTemplate(
                name=gen_valid_name("autogenerated-config"),
                project_id=project_id,
                config=config_object,
                anonymous=True,
            )
        ).result
        compute_template_id = str(created_template.id)
        return compute_template_id

    def _is_equal_cluster_compute(
        self, cluster_compute_1_id: str, cluster_compute_2_id: str,
    ) -> bool:
        """
        Compares config fields of two ComputeTemplate objects.
        """
        try:
            cluster_compute_1 = self.anyscale_api_client.get_compute_template(
                cluster_compute_1_id
            ).result.config
            cluster_compute_2 = self.anyscale_api_client.get_compute_template(
                cluster_compute_2_id
            ).result.config
            return bool(cluster_compute_1 == cluster_compute_2)
        except Exception as e:  # noqa: BLE001
            self.log.debug(f"Error comparing cluster compute: {e}")
            return False

    def _validate_new_cluster_compute_and_env_match_existing_cluster(
        self, project_id: str, running_cluster: Session, print_warnings: bool = True
    ) -> bool:
        """
        Compare the build id, compute template id, and allow_public_internet_traffic values
        for a existing running cluster and the values passed in. Print warnings if incompatible
        values if specified, and return if the specified arguments require a cluster restart.
        """
        update_required = False
        cluster_name = running_cluster.name
        if self.cluster_env_name:
            overriding_build = self._get_cluster_build(
                self.cluster_env_name, self.cluster_env_revision
            )
            if overriding_build.id != running_cluster.build_id:
                current_cluster_build = self.anyscale_api_client.get_build(
                    running_cluster.build_id
                ).result
                current_cluster_env = self.anyscale_api_client.get_app_config(
                    current_cluster_build.application_template_id
                ).result
                if print_warnings:
                    self.log.warning(
                        f"The cluster is currently using {current_cluster_env.name}:{current_cluster_build.revision} as the cluster env, "
                        f"yet {self.cluster_env_name}:{self.cluster_env_revision} was provided. The cluster will not be updated and will "
                        f"still use {current_cluster_env.name}:{current_cluster_build.revision} as the cluster env. To update the cluster, "
                        "please specify `update=True` in the anyscale address, e.g.: "
                        f'`ray.init("anyscale://{cluster_name}", cluster_env="{self.cluster_env_name}:{self.cluster_env_revision}", update=True)`'
                    )
                update_required = True
        if self.cluster_compute_name:
            # Because self.cluster_compute_name exists, this will get the id of an existing
            # cluster compute and not register a new one.
            overriding_compute = self._get_cluster_compute_id(
                project_id,
                self.cluster_compute_name,
                self.cluster_compute_dict,
                self.cloud_name,
            )
            if not self._is_equal_cluster_compute(
                overriding_compute, running_cluster.compute_template_id
            ):
                cluster_compute = self.anyscale_api_client.get_compute_template(
                    running_cluster.compute_template_id
                ).result
                if print_warnings:
                    self.log.warning(
                        f"The cluster is currently using {cluster_compute.name} as the compute config, "
                        f"yet {self.cluster_compute_name} was provided. The cluster will not be updated and will "
                        f"still use {cluster_compute.name} as the compute config. To update the cluster, please specify "
                        "`update=True` in the anyscale address, e.g.: "
                        f'`ray.init("anyscale://{cluster_name}", cluster_compute="{self.cluster_compute_name}", update=True")`'
                    )
                update_required = True
        if self.cluster_env_dict or self.cluster_compute_dict:
            self.log.warning(
                f"The cluster {cluster_name} is already active. To update the cluster please "
                "specify `update=True` in the anyscale address."
            )
            update_required = True

        if (
            self.allow_public_internet_traffic is not None
            and self.allow_public_internet_traffic
            != running_cluster.allow_public_internet_traffic
        ):
            if print_warnings:
                self.log.warning(
                    f"The cluster currently {'allows' if running_cluster.allow_public_internet_traffic else 'does not allow'} "
                    f"public internet traffic to the Serve endpoints, yet `allow_public_internet_traffic={self.allow_public_internet_traffic}` "
                    f"was specified. The cluster will not be updated and will continue {'allowing' if running_cluster.allow_public_internet_traffic else 'not allowing'} "
                    "public internet traffic to the Serve endpoints. To update the cluster, please specify `update=True` in the anyscale "
                    "address, e.g.: "
                    f'`ray.init("anyscale://{cluster_name}", update=True, allow_public_internet_traffic={self.allow_public_internet_traffic})`'
                )
            update_required = True

        return update_required


def create_prepare_cluster_block(  # noqa: PLR0913
    project_id: str,
    cluster_name: Optional[str],
    autosuspend_timeout: Optional[int],
    allow_public_internet_traffic: Optional[bool],
    needs_update: bool,
    cluster_compute_name: Optional[str],
    cluster_compute_dict: Optional[CLUSTER_COMPUTE_DICT_TYPE],
    cloud_name: Optional[str],
    build_pr: Optional[int],
    force_rebuild: bool,
    build_commit: Optional[str],
    cluster_env_name: Optional[str],
    cluster_env_dict: Optional[CLUSTER_ENV_DICT_TYPE],
    cluster_env_revision: Optional[int],
    ray: Any,
    log_output: bool = True,
):
    """
    Function to get PrepareClusterBlock object. The PrepareClusterBlock object
    is not a global variable an will be reinstantiated on each call to
    get_prepare_cluster_block.
    """
    prepare_cluster_block = PrepareClusterBlock(
        project_id=project_id,
        cluster_name=cluster_name,
        autosuspend_timeout=autosuspend_timeout,
        allow_public_internet_traffic=allow_public_internet_traffic,
        needs_update=needs_update,
        cluster_compute_name=cluster_compute_name,
        cluster_compute_dict=cluster_compute_dict,
        cloud_name=cloud_name,
        build_pr=build_pr,
        force_rebuild=force_rebuild,
        build_commit=build_commit,
        cluster_env_name=cluster_env_name,
        cluster_env_dict=cluster_env_dict,
        cluster_env_revision=cluster_env_revision,
        ray=ray,
        log_output=log_output,
    )
    prepare_cluster_block.prepare()
    return prepare_cluster_block



import click

from anyscale.controllers.llm.models_controller import ModelsController


@click.group("models", help="Finetuned models stored on your Anyscale cloud.")
def models_cli():
    pass


@models_cli.command(
    name="retrieve",
    short_help="Retrieve information for a model in your Anyscale cloud.",
)
@click.argument("model_id", required=True)
def retrieve_model(model_id: str) -> None:
    """
    Retrieves the model card for the given model ID.

    MODEL_ID = ID for the model of interest

    Example usage:
        anyscale llm models retrieve my-model-id
    """
    print(ModelsController().retrieve_model(model_id))



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedJobQueue(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'user_provided_id': 'str',
        'name': 'str',
        'current_job_queue_state': 'JobQueueState',
        'execution_mode': 'JobQueueExecutionMode',
        'max_concurrency': 'int',
        'idle_timeout_sec': 'int',
        'created_at': 'datetime',
        'creator_id': 'str',
        'creator_email': 'str',
        'compute_config_id': 'str',
        'current_cluster_state': 'SessionState',
        'cluster_id': 'str',
        'project_id': 'str',
        'total_jobs': 'int',
        'successful_jobs': 'int',
        'failed_jobs': 'int',
        'active_jobs': 'int'
    }

    attribute_map = {
        'id': 'id',
        'user_provided_id': 'user_provided_id',
        'name': 'name',
        'current_job_queue_state': 'current_job_queue_state',
        'execution_mode': 'execution_mode',
        'max_concurrency': 'max_concurrency',
        'idle_timeout_sec': 'idle_timeout_sec',
        'created_at': 'created_at',
        'creator_id': 'creator_id',
        'creator_email': 'creator_email',
        'compute_config_id': 'compute_config_id',
        'current_cluster_state': 'current_cluster_state',
        'cluster_id': 'cluster_id',
        'project_id': 'project_id',
        'total_jobs': 'total_jobs',
        'successful_jobs': 'successful_jobs',
        'failed_jobs': 'failed_jobs',
        'active_jobs': 'active_jobs'
    }

    def __init__(self, id=None, user_provided_id=None, name=None, current_job_queue_state=None, execution_mode=None, max_concurrency=None, idle_timeout_sec=None, created_at=None, creator_id=None, creator_email=None, compute_config_id=None, current_cluster_state=None, cluster_id=None, project_id=None, total_jobs=None, successful_jobs=None, failed_jobs=None, active_jobs=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedJobQueue - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._user_provided_id = None
        self._name = None
        self._current_job_queue_state = None
        self._execution_mode = None
        self._max_concurrency = None
        self._idle_timeout_sec = None
        self._created_at = None
        self._creator_id = None
        self._creator_email = None
        self._compute_config_id = None
        self._current_cluster_state = None
        self._cluster_id = None
        self._project_id = None
        self._total_jobs = None
        self._successful_jobs = None
        self._failed_jobs = None
        self._active_jobs = None
        self.discriminator = None

        self.id = id
        self.user_provided_id = user_provided_id
        self.name = name
        self.current_job_queue_state = current_job_queue_state
        self.execution_mode = execution_mode
        self.max_concurrency = max_concurrency
        self.idle_timeout_sec = idle_timeout_sec
        self.created_at = created_at
        self.creator_id = creator_id
        self.creator_email = creator_email
        self.compute_config_id = compute_config_id
        if current_cluster_state is not None:
            self.current_cluster_state = current_cluster_state
        if cluster_id is not None:
            self.cluster_id = cluster_id
        self.project_id = project_id
        if total_jobs is not None:
            self.total_jobs = total_jobs
        if successful_jobs is not None:
            self.successful_jobs = successful_jobs
        if failed_jobs is not None:
            self.failed_jobs = failed_jobs
        if active_jobs is not None:
            self.active_jobs = active_jobs

    @property
    def id(self):
        """Gets the id of this DecoratedJobQueue.  # noqa: E501

        Server assigned unique identifier of the job queue.  # noqa: E501

        :return: The id of this DecoratedJobQueue.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedJobQueue.

        Server assigned unique identifier of the job queue.  # noqa: E501

        :param id: The id of this DecoratedJobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def user_provided_id(self):
        """Gets the user_provided_id of this DecoratedJobQueue.  # noqa: E501

        User provided identifier of the job queue.  # noqa: E501

        :return: The user_provided_id of this DecoratedJobQueue.  # noqa: E501
        :rtype: str
        """
        return self._user_provided_id

    @user_provided_id.setter
    def user_provided_id(self, user_provided_id):
        """Sets the user_provided_id of this DecoratedJobQueue.

        User provided identifier of the job queue.  # noqa: E501

        :param user_provided_id: The user_provided_id of this DecoratedJobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user_provided_id is None:  # noqa: E501
            raise ValueError("Invalid value for `user_provided_id`, must not be `None`")  # noqa: E501

        self._user_provided_id = user_provided_id

    @property
    def name(self):
        """Gets the name of this DecoratedJobQueue.  # noqa: E501

        Name of the job queue. Drived from the field user_provided_id.  # noqa: E501

        :return: The name of this DecoratedJobQueue.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this DecoratedJobQueue.

        Name of the job queue. Drived from the field user_provided_id.  # noqa: E501

        :param name: The name of this DecoratedJobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def current_job_queue_state(self):
        """Gets the current_job_queue_state of this DecoratedJobQueue.  # noqa: E501

        The current state of the job queue.  # noqa: E501

        :return: The current_job_queue_state of this DecoratedJobQueue.  # noqa: E501
        :rtype: JobQueueState
        """
        return self._current_job_queue_state

    @current_job_queue_state.setter
    def current_job_queue_state(self, current_job_queue_state):
        """Sets the current_job_queue_state of this DecoratedJobQueue.

        The current state of the job queue.  # noqa: E501

        :param current_job_queue_state: The current_job_queue_state of this DecoratedJobQueue.  # noqa: E501
        :type: JobQueueState
        """
        if self.local_vars_configuration.client_side_validation and current_job_queue_state is None:  # noqa: E501
            raise ValueError("Invalid value for `current_job_queue_state`, must not be `None`")  # noqa: E501

        self._current_job_queue_state = current_job_queue_state

    @property
    def execution_mode(self):
        """Gets the execution_mode of this DecoratedJobQueue.  # noqa: E501

        The execution mode of the job queue.  # noqa: E501

        :return: The execution_mode of this DecoratedJobQueue.  # noqa: E501
        :rtype: JobQueueExecutionMode
        """
        return self._execution_mode

    @execution_mode.setter
    def execution_mode(self, execution_mode):
        """Sets the execution_mode of this DecoratedJobQueue.

        The execution mode of the job queue.  # noqa: E501

        :param execution_mode: The execution_mode of this DecoratedJobQueue.  # noqa: E501
        :type: JobQueueExecutionMode
        """
        if self.local_vars_configuration.client_side_validation and execution_mode is None:  # noqa: E501
            raise ValueError("Invalid value for `execution_mode`, must not be `None`")  # noqa: E501

        self._execution_mode = execution_mode

    @property
    def max_concurrency(self):
        """Gets the max_concurrency of this DecoratedJobQueue.  # noqa: E501

        The maximum number of jobs that can be run concurrently.  # noqa: E501

        :return: The max_concurrency of this DecoratedJobQueue.  # noqa: E501
        :rtype: int
        """
        return self._max_concurrency

    @max_concurrency.setter
    def max_concurrency(self, max_concurrency):
        """Sets the max_concurrency of this DecoratedJobQueue.

        The maximum number of jobs that can be run concurrently.  # noqa: E501

        :param max_concurrency: The max_concurrency of this DecoratedJobQueue.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and max_concurrency is None:  # noqa: E501
            raise ValueError("Invalid value for `max_concurrency`, must not be `None`")  # noqa: E501

        self._max_concurrency = max_concurrency

    @property
    def idle_timeout_sec(self):
        """Gets the idle_timeout_sec of this DecoratedJobQueue.  # noqa: E501

        The number of seconds a job can be idle before it is terminated.  # noqa: E501

        :return: The idle_timeout_sec of this DecoratedJobQueue.  # noqa: E501
        :rtype: int
        """
        return self._idle_timeout_sec

    @idle_timeout_sec.setter
    def idle_timeout_sec(self, idle_timeout_sec):
        """Sets the idle_timeout_sec of this DecoratedJobQueue.

        The number of seconds a job can be idle before it is terminated.  # noqa: E501

        :param idle_timeout_sec: The idle_timeout_sec of this DecoratedJobQueue.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and idle_timeout_sec is None:  # noqa: E501
            raise ValueError("Invalid value for `idle_timeout_sec`, must not be `None`")  # noqa: E501

        self._idle_timeout_sec = idle_timeout_sec

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedJobQueue.  # noqa: E501

        Time at which job queue was created.  # noqa: E501

        :return: The created_at of this DecoratedJobQueue.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedJobQueue.

        Time at which job queue was created.  # noqa: E501

        :param created_at: The created_at of this DecoratedJobQueue.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def creator_id(self):
        """Gets the creator_id of this DecoratedJobQueue.  # noqa: E501

        Identifier of user who created the job queue.  # noqa: E501

        :return: The creator_id of this DecoratedJobQueue.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this DecoratedJobQueue.

        Identifier of user who created the job queue.  # noqa: E501

        :param creator_id: The creator_id of this DecoratedJobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def creator_email(self):
        """Gets the creator_email of this DecoratedJobQueue.  # noqa: E501

        Email of user who created the job queue.  # noqa: E501

        :return: The creator_email of this DecoratedJobQueue.  # noqa: E501
        :rtype: str
        """
        return self._creator_email

    @creator_email.setter
    def creator_email(self, creator_email):
        """Sets the creator_email of this DecoratedJobQueue.

        Email of user who created the job queue.  # noqa: E501

        :param creator_email: The creator_email of this DecoratedJobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_email is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_email`, must not be `None`")  # noqa: E501

        self._creator_email = creator_email

    @property
    def compute_config_id(self):
        """Gets the compute_config_id of this DecoratedJobQueue.  # noqa: E501

        The compute config id for the job queue  # noqa: E501

        :return: The compute_config_id of this DecoratedJobQueue.  # noqa: E501
        :rtype: str
        """
        return self._compute_config_id

    @compute_config_id.setter
    def compute_config_id(self, compute_config_id):
        """Sets the compute_config_id of this DecoratedJobQueue.

        The compute config id for the job queue  # noqa: E501

        :param compute_config_id: The compute_config_id of this DecoratedJobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and compute_config_id is None:  # noqa: E501
            raise ValueError("Invalid value for `compute_config_id`, must not be `None`")  # noqa: E501

        self._compute_config_id = compute_config_id

    @property
    def current_cluster_state(self):
        """Gets the current_cluster_state of this DecoratedJobQueue.  # noqa: E501

        The current state of the cluster associated with the job queue.  # noqa: E501

        :return: The current_cluster_state of this DecoratedJobQueue.  # noqa: E501
        :rtype: SessionState
        """
        return self._current_cluster_state

    @current_cluster_state.setter
    def current_cluster_state(self, current_cluster_state):
        """Sets the current_cluster_state of this DecoratedJobQueue.

        The current state of the cluster associated with the job queue.  # noqa: E501

        :param current_cluster_state: The current_cluster_state of this DecoratedJobQueue.  # noqa: E501
        :type: SessionState
        """

        self._current_cluster_state = current_cluster_state

    @property
    def cluster_id(self):
        """Gets the cluster_id of this DecoratedJobQueue.  # noqa: E501

        The cluster id associated with the job queue. Cluster id could be null, in case when Job Queue has been created but cluster has not been provisioned yet.  # noqa: E501

        :return: The cluster_id of this DecoratedJobQueue.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this DecoratedJobQueue.

        The cluster id associated with the job queue. Cluster id could be null, in case when Job Queue has been created but cluster has not been provisioned yet.  # noqa: E501

        :param cluster_id: The cluster_id of this DecoratedJobQueue.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def project_id(self):
        """Gets the project_id of this DecoratedJobQueue.  # noqa: E501

        The project id associated with the job queue.  # noqa: E501

        :return: The project_id of this DecoratedJobQueue.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this DecoratedJobQueue.

        The project id associated with the job queue.  # noqa: E501

        :param project_id: The project_id of this DecoratedJobQueue.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def total_jobs(self):
        """Gets the total_jobs of this DecoratedJobQueue.  # noqa: E501

        Total number of jobs in the job queue.  # noqa: E501

        :return: The total_jobs of this DecoratedJobQueue.  # noqa: E501
        :rtype: int
        """
        return self._total_jobs

    @total_jobs.setter
    def total_jobs(self, total_jobs):
        """Sets the total_jobs of this DecoratedJobQueue.

        Total number of jobs in the job queue.  # noqa: E501

        :param total_jobs: The total_jobs of this DecoratedJobQueue.  # noqa: E501
        :type: int
        """

        self._total_jobs = total_jobs

    @property
    def successful_jobs(self):
        """Gets the successful_jobs of this DecoratedJobQueue.  # noqa: E501

        Number of successful jobs in the job queue.  # noqa: E501

        :return: The successful_jobs of this DecoratedJobQueue.  # noqa: E501
        :rtype: int
        """
        return self._successful_jobs

    @successful_jobs.setter
    def successful_jobs(self, successful_jobs):
        """Sets the successful_jobs of this DecoratedJobQueue.

        Number of successful jobs in the job queue.  # noqa: E501

        :param successful_jobs: The successful_jobs of this DecoratedJobQueue.  # noqa: E501
        :type: int
        """

        self._successful_jobs = successful_jobs

    @property
    def failed_jobs(self):
        """Gets the failed_jobs of this DecoratedJobQueue.  # noqa: E501

        Number of failed jobs in the job queue.  # noqa: E501

        :return: The failed_jobs of this DecoratedJobQueue.  # noqa: E501
        :rtype: int
        """
        return self._failed_jobs

    @failed_jobs.setter
    def failed_jobs(self, failed_jobs):
        """Sets the failed_jobs of this DecoratedJobQueue.

        Number of failed jobs in the job queue.  # noqa: E501

        :param failed_jobs: The failed_jobs of this DecoratedJobQueue.  # noqa: E501
        :type: int
        """

        self._failed_jobs = failed_jobs

    @property
    def active_jobs(self):
        """Gets the active_jobs of this DecoratedJobQueue.  # noqa: E501

        Number of active jobs in the job queue.  # noqa: E501

        :return: The active_jobs of this DecoratedJobQueue.  # noqa: E501
        :rtype: int
        """
        return self._active_jobs

    @active_jobs.setter
    def active_jobs(self, active_jobs):
        """Sets the active_jobs of this DecoratedJobQueue.

        Number of active jobs in the job queue.  # noqa: E501

        :param active_jobs: The active_jobs of this DecoratedJobQueue.  # noqa: E501
        :type: int
        """

        self._active_jobs = active_jobs

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedJobQueue):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedJobQueue):
            return True

        return self.to_dict() != other.to_dict()



from typing import Any, Dict, Optional, Tuple

from anyscale._private.sdk import sdk_command
from anyscale.workspace._private.workspace_sdk import PrivateWorkspaceSDK
from anyscale.workspace.models import WorkspaceConfig, WorkspaceState


_WORKSPACE_SDK_SINGLETON_KEY = "workspace_sdk"

_CREATE_EXAMPLE = """
import anyscale
from anyscale.workspace.models import WorkspaceConfig

anyscale.workspace.create(
    WorkspaceConfig(
        name="my-workspace",
        idle_termination_minutes=120,
    ),
)
"""

_CREATE_ARG_DOCSTRINGS = {"config": "The config for defining the workspace."}
_WAIT_TIMEOUT_SECONDS = 1800.0


@sdk_command(
    _WORKSPACE_SDK_SINGLETON_KEY,
    PrivateWorkspaceSDK,
    doc_py_example=_CREATE_EXAMPLE,
    arg_docstrings=_CREATE_ARG_DOCSTRINGS,
)
def create(config: WorkspaceConfig, *, _sdk: PrivateWorkspaceSDK) -> str:
    """Create a workspace.

    Returns the id of the created workspace.
    """
    return _sdk.create(config)


_START_EXAMPLE = """
import anyscale

anyscale.workspace.start(
    name="my-workspace",
)
"""

_START_ARG_DOCSTRINGS = {
    "name": "Name of the workspace.",
    "id": "Unique ID of the workspace",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _WORKSPACE_SDK_SINGLETON_KEY,
    PrivateWorkspaceSDK,
    doc_py_example=_START_EXAMPLE,
    arg_docstrings=_START_ARG_DOCSTRINGS,
)
def start(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _sdk: PrivateWorkspaceSDK,
) -> str:
    """Start a workspace.

    Returns the id of the started workspace.
    """
    return _sdk.start(name=name, id=id, cloud=cloud, project=project)


_TERMINATE_EXAMPLE = """
import anyscale

anyscale.workspace.terminate(
    name="my-workspace",
)
"""

_TERMINATE_ARG_DOCSTRINGS = {
    "name": "Name of the workspace.",
    "id": "Unique ID of the workspace",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _WORKSPACE_SDK_SINGLETON_KEY,
    PrivateWorkspaceSDK,
    doc_py_example=_TERMINATE_EXAMPLE,
    arg_docstrings=_TERMINATE_ARG_DOCSTRINGS,
)
def terminate(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _sdk: PrivateWorkspaceSDK,
) -> str:
    """Terminate a workspace.

    Returns the id of the terminated workspace.
    """
    return _sdk.terminate(name=name, id=id, cloud=cloud, project=project)


_STATUS_EXAMPLE = """
import anyscale

status = anyscale.workspace.status(
    name="my-workspace",
)
"""

_STATUS_ARG_DOCSTRINGS = {
    "name": "Name of the workspace.",
    "id": "Unique ID of the workspace",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _WORKSPACE_SDK_SINGLETON_KEY,
    PrivateWorkspaceSDK,
    doc_py_example=_STATUS_EXAMPLE,
    arg_docstrings=_STATUS_ARG_DOCSTRINGS,
)
def status(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _sdk: PrivateWorkspaceSDK,
) -> str:
    """Get the status of a workspace.

    Returns the status of the workspace.
    """
    return _sdk.status(name=name, id=id, cloud=cloud, project=project)


_WAIT_EXAMPLE = """
import anyscale

anyscale.workspace.wait(
    name="my-workspace",
)
"""

_WAIT_ARG_DOCSTRINGS = {
    "name": "Name of the workspace.",
    "id": "Unique ID of the workspace",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
    "timeout_s": "The maximum time to wait for the workspace to reach a terminal state.",
    "state": "The desired terminal state to wait for, defaults to RUNNING.",
}


@sdk_command(
    _WORKSPACE_SDK_SINGLETON_KEY,
    PrivateWorkspaceSDK,
    doc_py_example=_WAIT_EXAMPLE,
    arg_docstrings=_WAIT_ARG_DOCSTRINGS,
)
def wait(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    timeout_s: float = _WAIT_TIMEOUT_SECONDS,
    state: str = WorkspaceState.RUNNING,
    _sdk: PrivateWorkspaceSDK,
) -> str:
    """Wait for a workspace to reach a terminal state.

    Returns the status of the workspace.
    """
    return _sdk.wait(
        name=name,
        id=id,
        cloud=cloud,
        project=project,
        timeout_s=timeout_s,
        state=state,
    )


_GENERATE_SSH_CONFIG_FILE_EXAMPLE = """
import anyscale
import subprocess

host_name, config_file = anyscale.workspace.generate_ssh_config_file(
    name="my-workspace",
)

# run an ssh command using the generated config file
subprocess.run(["ssh", "-F", config_path, host_name, "ray --version"])
"""

_GENERATE_SSH_CONFIG_FILE_ARG_DOCSTRINGS = {
    "name": "Name of the workspace.",
    "id": "Unique ID of the workspace",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
    "ssh_config_path": "The directory to write the generated config file to.",
}


@sdk_command(
    _WORKSPACE_SDK_SINGLETON_KEY,
    PrivateWorkspaceSDK,
    doc_py_example=_GENERATE_SSH_CONFIG_FILE_EXAMPLE,
    arg_docstrings=_GENERATE_SSH_CONFIG_FILE_ARG_DOCSTRINGS,
)
def generate_ssh_config_file(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    ssh_config_path: Optional[str] = None,
    _sdk: PrivateWorkspaceSDK,
) -> Tuple[str, str]:
    """Generate an SSH config file for a workspace.

    Returns the hostname and path to the generated config file.
    """
    return _sdk.generate_ssh_config_file(
        name=name, id=id, cloud=cloud, project=project, ssh_config_path=ssh_config_path,
    )


_RUN_COMMAND_EXAMPLE = """
import anyscale

process = anyscale.workspace.run_command(
    name="my-workspace",
    command="ray_version",
    capture_output=True,
    text=True,
)
print(process.stdout)
"""

_RUN_COMMAND_ARG_DOCSTRINGS = {
    "name": "Name of the workspace.",
    "id": "Unique ID of the workspace",
    "cloud": "The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
    "command": "The command to run.",
    "kwargs": "Additional arguments to pass to subprocess.run.",
}


@sdk_command(
    _WORKSPACE_SDK_SINGLETON_KEY,
    PrivateWorkspaceSDK,
    doc_py_example=_RUN_COMMAND_EXAMPLE,
    arg_docstrings=_RUN_COMMAND_ARG_DOCSTRINGS,
)
def run_command(
    *,
    name: Optional[str] = None,
    id: Optional[str] = None,  # noqa: A002
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    command: str,
    _sdk: PrivateWorkspaceSDK,
    **kwargs: Dict[str, Any],
):
    """Run a command in a workspace.

    Returns a subprocess.CompletedProcess object.
    """
    return _sdk.run_command(
        name=name, id=id, cloud=cloud, project=project, command=command, **kwargs,
    )



import json
import os
import re
import time
from typing import Any, Dict
from unittest.mock import Mock, patch

import boto3
from botocore.exceptions import ClientError
from click import ClickException
from moto import mock_ec2
from packaging import version
import pytest

from anyscale.aws_iam_policies import (
    ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE,
    DEFAULT_RAY_IAM_ASSUME_ROLE_POLICY,
    get_anyscale_aws_iam_assume_role_policy,
)
from anyscale.client.openapi_client.models import CloudAnalyticsEventCloudResource
from anyscale.conf import MINIMUM_RAY_VERSION
from anyscale.util import (
    _check_python_version,
    _get_memorydb_cluster_config,
    _get_subnet,
    _ray_version_major_minor,
    _update_external_ids_for_policy,
    contains_control_plane_role,
    DEFAULT_RAY_VERSION,
    extract_versions_from_image_name,
    filter_actions_from_policy_document,
    get_cluster_model_for_current_workspace,
    get_latest_ray_version,
    get_ray_and_py_version_for_default_cluster_env,
    poll,
    populate_unspecified_cluster_configs,
    prepare_cloudformation_template,
    sleep_till,
    str_data_size,
    updating_printer,
    verify_data_plane_role_assume_role_policy,
)
from anyscale.utils.cloud_utils import CloudSetupError
from anyscale.utils.name_utils import gen_valid_name


# Keep in sync with backend.server.common.utils.constants.NAME_VALIDATION_REGEX_PATTERN
_NAME_VALIDATION_REGEX_PATTERN = "^[A-Za-z0-9._-]+$"


def test_updating_printer() -> None:
    out = ""

    def mock_print(
        string: str, *args: Any, end: str = "\n", flush: bool = False, **kwargs: Any
    ) -> None:
        nonlocal out
        out += string
        out += end

    with patch("anyscale.util.print", new=mock_print), patch(
        "shutil.get_terminal_size"
    ) as get_terminal_size_mock:
        get_terminal_size_mock.return_value = (10, 24)
        with updating_printer() as print_status:
            print_status("Step 1")
            print_status("Step 2")
            print_status("Step 3")

    assert out == (
        "\r          \r"
        "Step 1"
        "\r          \r"
        "Step 2"
        "\r          \r"
        "Step 3"
        "\r          \r"
    )


def test_updating_printer_multiline() -> None:
    out = ""

    def mock_print(
        string: str, *args: Any, end: str = "\n", flush: bool = False, **kwargs: Any
    ) -> None:
        nonlocal out
        out += string
        out += end

    with patch("anyscale.util.print", new=mock_print), patch(
        "shutil.get_terminal_size"
    ) as get_terminal_size_mock:
        get_terminal_size_mock.return_value = (10, 24)
        with updating_printer() as print_status:
            print_status("Step 1\nExtra stuff")
            print_status("ExtraLongLine12345")
            print_status("ExtraLongLine12345\nExtra stuff")
            print_status("Step 3")

    assert out == (
        "\r          \r"
        "Step 1..."
        "\r          \r"
        "ExtraLo..."
        "\r          \r"
        "ExtraLo..."
        "\r          \r"
        "Step 3"
        "\r          \r"
    )


STATEMENT_TEMPLATE = {
    "Action": "sts:AssumeRole",
    "Effect": "Allow",
    "Principal": {"AWS": "arn:aws:iam::ACCT_ID:root"},
}


@pytest.mark.parametrize(
    ("statement_policy", "expected_conditions"),
    [
        pytest.param(
            [STATEMENT_TEMPLATE],
            [{"StringEquals": {"sts:ExternalId": ["new_id"]}}],
            id="OneStatement,NoPrior",
        ),
        pytest.param(
            [STATEMENT_TEMPLATE, STATEMENT_TEMPLATE],
            [{"StringEquals": {"sts:ExternalId": ["new_id"]}}] * 2,
            id="TwoStatements,NoPrior",
        ),
        pytest.param(
            [
                {
                    "Condition": {"StringEquals": {"sts:ExternalId": "old_id"}},
                    **STATEMENT_TEMPLATE,  # type: ignore
                }
            ],
            [{"StringEquals": {"sts:ExternalId": ["old_id", "new_id"]}}],
            id="OneStatement,OnePriorExternal",
        ),
        pytest.param(
            [
                {
                    "Condition": {"StringEquals": {"sts:ExternalId": "old_id"}},
                    **STATEMENT_TEMPLATE,  # type: ignore
                },
                STATEMENT_TEMPLATE,
            ],
            [
                {"StringEquals": {"sts:ExternalId": ["old_id", "new_id"]}},
                {"StringEquals": {"sts:ExternalId": ["new_id"]}},
            ],
            id="TwoStatements,OnePriorExternal",
        ),
        pytest.param(
            [
                {
                    "Condition": {"StringNotEquals": {"sts:ExternalId": "old_id"}},
                    **STATEMENT_TEMPLATE,  # type: ignore
                },
                STATEMENT_TEMPLATE,
            ],
            [
                {
                    "StringEquals": {"sts:ExternalId": ["new_id"]},
                    "StringNotEquals": {"sts:ExternalId": "old_id"},
                },
            ],
            id="OneStatemnt,OtherCondition",
        ),
    ],
)
def test_update_external_ids_for_policy(statement_policy, expected_conditions):
    policy_document = {
        "Statement": statement_policy,
        "Version": "2012-10-17",
    }
    new_policy = _update_external_ids_for_policy(policy_document, "new_id")

    for new, expected in zip(new_policy["Statement"], expected_conditions):
        assert new["Condition"] == expected


@pytest.mark.parametrize(
    ("image_name", "expected", "exception_substr"),
    [
        ("anyscale/ray-ml:1.11.1-py38-gpu", ("py38", "1.11.1"), None),
        ("anyscale/ray:1.12-py37-cpu", ("py37", "1.12"), None),
        ("anyscale/ray:1.12-py37", ("py37", "1.12"), None),
        ("anyscale/ray:1.12py37", None, "got 1.12py37"),
    ],
)
def test_extract_versions_from_image_name(image_name, expected, exception_substr):
    if exception_substr is not None:
        with pytest.raises(ValueError) as exc_info:
            extract_versions_from_image_name(image_name)
        assert exception_substr in str(exc_info.value)
    else:
        python_version, ray_version = extract_versions_from_image_name(image_name)
        assert (python_version, ray_version) == expected


@pytest.mark.parametrize(
    ("ray_version", "expected", "exception_substr"),
    [
        ("1.12", (1, 12), None),
        ("0.0", (0, 0), None),
        ("0:0", (0, 0), "unexpected"),
        ("112", (0, 0), "unexpected"),
        ("", (0, 0), "unexpected"),
        ("1.x", (0, 0), "unexpected"),
        ("0x10.12", (0, 0), "unexpected"),
    ],
)
def test_ray_version_major_minor(ray_version, expected, exception_substr):
    if exception_substr is not None:
        with pytest.raises(Exception) as exc_info:  # noqa: PT011
            _ray_version_major_minor(ray_version)
        assert exception_substr in str(exc_info.value)
    else:
        got = _ray_version_major_minor(ray_version)
        assert got == expected


@pytest.mark.parametrize(
    ("python_version", "exception_substr"),
    [
        ("py36", None),
        ("py37", None),
        ("py38", None),
        ("py39", None),
        ("py10", "got py10"),
        ("py3.6", "got py3.6"),
        ("py3", "got py3."),
        ("py35", "got py35"),
    ],
)
def test_python_version_major_minor(python_version, exception_substr):
    if exception_substr is not None:
        with pytest.raises(Exception) as exc_info:  # noqa: PT011
            _check_python_version(python_version)
        assert exception_substr in str(exc_info.value)
    else:
        _check_python_version(python_version)


def test_poll():
    """Test the poll function."""

    end_time = time.time() + 0.5
    sleep_till(end_time)
    assert time.time() == pytest.approx(end_time)

    # This should poll forever
    count = 0
    start_time = time.time()
    for i in poll(0.01):
        count += 1
        assert count == i
        if count > 100:
            break
    assert count == 101
    assert time.time() == pytest.approx(start_time + 1.01)

    # Assert we stop iterating at max iter
    expected_i = 1
    for i in poll(0.01, max_iter=5):
        assert i == expected_i
        expected_i += 1
        assert i <= 5


def test_str_data_size():
    assert str_data_size("abcd") == 4

    # Serialized form: '{"hi": "ih"}'
    assert str_data_size(json.dumps({"hi": "ih"})) == 12


@pytest.mark.parametrize("is_anyscale_hosted", [True, False])
def test_prepare_cloudformation_template(is_anyscale_hosted):

    mock_region = "us-east-1"
    mock_cfn_stack_name = "mock_cfn_stack_name"
    mock_cloud_id = "mock_cloud_id"

    mock_azs = {
        "use1-az1": "us-east-1a",
        "use1-az2": "us-east-1b",
        "use1-az3": "us-east-1c",
        "use1-az4": "us-east-1d",
        "use1-az5": "us-east-1e",
        "use1-az6": "us-east-1f",
    }

    with patch("anyscale.util.get_availability_zones", new=Mock(return_value=mock_azs)):
        cfn_template = prepare_cloudformation_template(
            mock_region,
            mock_cfn_stack_name,
            mock_cloud_id,
            True,
            is_anyscale_hosted=is_anyscale_hosted,
        )

    # If you make some changes to the prepare_cloudformation_template(), update the
    # cloud_formation_template.txt by commenting out these code. And use `bazelisk run`
    # to update the template (bazelisk run will not use the sandbox so that we have the
    # write permission).
    # with open(f"{os.environ['HOME']}/product/frontend/cli/tests/cloud_formation_template.txt", 'w') as f:
    #     f.write(cfn_template)
    cft_name = "cloud_formation_template.txt"
    if is_anyscale_hosted:
        cft_name = "cloud_formation_template_oa.txt"
    with open(f"{os.path.dirname(os.path.abspath(__file__))}/{cft_name}") as f:
        expected_content = f.read()
    assert cfn_template == expected_content


@pytest.mark.parametrize("request_output", ["mock_latest_version", Exception()])
def test_get_latest_ray_version(request_output):
    if isinstance(request_output, str):
        mock_get = Mock(
            return_value=Mock(
                json=Mock(return_value={"info": {"version": request_output}})
            )
        )
    else:
        mock_get = Mock(side_effect=request_output)

    expected_latest_version = (
        request_output if isinstance(request_output, str) else DEFAULT_RAY_VERSION
    )
    with patch.multiple("anyscale.util.requests", get=mock_get):
        latest_ray_version = get_latest_ray_version()
        assert latest_ray_version == expected_latest_version
    mock_get.assert_called_once_with("https://pypi.org/pypi/ray/json")


@pytest.mark.parametrize(
    "pyversion_param", [["3", "6"], ["3", "7"], ["3", "8"], ["3", "9"]]
)
def test_get_ray_and_py_version_for_default_cluster_env(pyversion_param):
    # TODO(nikita): This test should be run in an environment that has Ray
    # installed, and one that doesn't have Ray installed. The
    # `import ray` statement inside the method cannot be mocked.
    mock_get_latest_ray_version = Mock(return_value="mock_latest_ray_version")
    with patch.multiple(
        "anyscale.util.sys", version_info=pyversion_param
    ), patch.multiple(
        "anyscale.util", get_latest_ray_version=mock_get_latest_ray_version
    ):
        ray_version, pyversion = get_ray_and_py_version_for_default_cluster_env()
    assert pyversion == "".join(str(x) for x in pyversion_param)
    try:
        import ray

        assert version.parse(ray_version) == version.parse(ray.__version__)
        assert version.parse(ray_version) >= version.parse(MINIMUM_RAY_VERSION)
    except ImportError:
        assert ray_version == "mock_latest_ray_version"


test_workspace_cluster = Mock()
test_workspace_cluster.name = "workspace-cluster-test_name"
test_workspace_cluster.project_id = "test_project_id"
test_workspace_cluster.cluster_environment_build_id = "test_build_id"
test_workspace_cluster.cluster_compute_id = "test_compute_config_id"


def test_get_cluster_model_for_current_workspace():
    api_client_result_mock = Mock()
    api_client_result_mock.result = test_workspace_cluster

    mock_base_api_client = Mock()
    mock_base_api_client.get_cluster = Mock(return_value=api_client_result_mock)

    # If workspace_id or session_id is missing in the env, should return None.
    assert get_cluster_model_for_current_workspace(mock_base_api_client) is None
    mock_base_api_client.get_cluster.assert_not_called()

    with patch.dict(
        os.environ,
        {"ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": "test_workspace_id",},
        clear=True,
    ):
        assert get_cluster_model_for_current_workspace(mock_base_api_client) is None
        mock_base_api_client.get_cluster.assert_not_called()

    with patch.dict(
        os.environ, {"ANYSCALE_SESSION_ID": "test_session_id",}, clear=True,
    ):
        assert get_cluster_model_for_current_workspace(mock_base_api_client) is None
        mock_base_api_client.get_cluster.assert_not_called()

    # If both workspace_id and session_id are present in the env, should call the API
    # and return the cluster model.
    with patch.dict(
        os.environ,
        {
            "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID": "test_workspace_id",
            "ANYSCALE_SESSION_ID": "test_session_id",
        },
        clear=True,
    ):
        assert (
            get_cluster_model_for_current_workspace(mock_base_api_client)
            == test_workspace_cluster
        )
        mock_base_api_client.get_cluster.assert_called_once_with("test_session_id")


@pytest.mark.parametrize(
    ("input_config", "output_config"),
    [
        # Test all fields being defaulted.
        (
            {"name": "test_job"},
            {
                "name": "test_job",
                "project_id": "test_project_id",
                "build_id": "test_build_id",
                "compute_config_id": "test_compute_config_id",
            },
        ),
        # Test no fields being defaulted.
        (
            {
                "name": "test_job",
                "project_id": "existing_project_id",
                "build_id": "existing_build_id",
                "compute_config_id": "existing_compute_config_id",
            },
            {
                "name": "test_job",
                "project_id": "existing_project_id",
                "build_id": "existing_build_id",
                "compute_config_id": "existing_compute_config_id",
            },
        ),
        # If project_id is specified, it shouldn't be overwritten.
        (
            {"name": "test_job", "project_id": "existing_project_id"},
            {
                "name": "test_job",
                "project_id": "existing_project_id",
                "build_id": "test_build_id",
                "compute_config_id": "test_compute_config_id",
            },
        ),
        # If cluster_env or build_id are specified, they shouldn't be changed.
        (
            {"name": "test_job", "cluster_env": "existing_cluster_env"},
            {
                "name": "test_job",
                "project_id": "test_project_id",
                "cluster_env": "existing_cluster_env",
                "compute_config_id": "test_compute_config_id",
            },
        ),
        (
            {"name": "test_job", "build_id": "existing_build_id"},
            {
                "name": "test_job",
                "project_id": "test_project_id",
                "build_id": "existing_build_id",
                "compute_config_id": "test_compute_config_id",
            },
        ),
        # If cloud, compute_config, or compute_config_id are specified, they shouldn't be changed.
        (
            {"name": "test_job", "cloud": "test_cloud"},
            {
                "name": "test_job",
                "project_id": "test_project_id",
                "build_id": "test_build_id",
                "cloud": "test_cloud",
            },
        ),
        (
            {"name": "test_job", "compute_config": "test_compute_config"},
            {
                "name": "test_job",
                "project_id": "test_project_id",
                "build_id": "test_build_id",
                "compute_config": "test_compute_config",
            },
        ),
        (
            {"name": "test_job", "compute_config_id": "existing_compute_config_id"},
            {
                "name": "test_job",
                "project_id": "test_project_id",
                "build_id": "test_build_id",
                "compute_config_id": "existing_compute_config_id",
            },
        ),
    ],
)
def test_populate_unspecified_cluster_configs(
    input_config: Dict[str, str], output_config: Dict[str, str]
):
    assert (
        populate_unspecified_cluster_configs(input_config, test_workspace_cluster)
        == output_config
    )


def test_populate_unspecified_cluster_configs_populate_name():
    base_expected_output = {
        "project_id": "test_project_id",
        "build_id": "test_build_id",
        "compute_config_id": "test_compute_config_id",
    }

    # If a name is passed in, it shouldn't be overwritten.
    assert populate_unspecified_cluster_configs(
        {"name": "existing_name"}, test_workspace_cluster, populate_name=True
    ) == {"name": "existing_name", **base_expected_output}

    # If a name is not passed in and populate_name=False, it shouldn't be overwritten.
    assert populate_unspecified_cluster_configs({}, test_workspace_cluster,) == {
        **base_expected_output
    }

    # If a name is not passed in, it should default to the cluster name stripped of the workspace prefix.
    assert populate_unspecified_cluster_configs(
        {}, test_workspace_cluster, populate_name=True
    ) == {"name": "test_name", **base_expected_output}

    # That that missing the workspace prefix in the cluster name picks it up as-is.
    test_workspace_cluster_no_name_prefix = Mock()
    test_workspace_cluster_no_name_prefix.name = "oopsies_unexpected_name"
    test_workspace_cluster_no_name_prefix.project_id = "test_project_id"
    test_workspace_cluster_no_name_prefix.cluster_environment_build_id = "test_build_id"
    test_workspace_cluster_no_name_prefix.cluster_compute_id = "test_compute_config_id"
    assert populate_unspecified_cluster_configs(
        {}, test_workspace_cluster_no_name_prefix, populate_name=True
    ) == {"name": "oopsies_unexpected_name", **base_expected_output}


@pytest.mark.parametrize(
    ("policy_document", "expected_actions"),
    [
        pytest.param(
            ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE,
            {
                "iam:PassRole",
                "iam:CreateServiceLinkedRole",
                "iam:GetInstanceProfile",
                "ec2:DescribeAvailabilityZones",
                "ec2:DescribeInstanceTypes",
                "ec2:DescribeRegions",
                "ec2:DescribeAccountAttributes",
                "ec2:DescribeInstances",
                "ec2:DescribeSubnets",
                "ec2:DescribeRouteTables",
                "ec2:DescribeSecurityGroups",
                "ec2:RunInstances",
                "ec2:StartInstances",
                "ec2:StopInstances",
                "ec2:TerminateInstances",
                "ec2:CreateTags",
                "ec2:DeleteTags",
                "ec2:CancelSpotInstanceRequests",
                "ec2:ModifyImageAttribute",
                "ec2:ModifyInstanceAttribute",
                "ec2:RequestSpotInstances",
                "ec2:AttachVolume",
                "ec2:CreateVolume",
                "ec2:DescribeVolumes",
                "ec2:AssociateIamInstanceProfile",
                "ec2:DisassociateIamInstanceProfile",
                "ec2:ReplaceIamInstanceProfileAssociation",
                "ec2:CreatePlacementGroup",
                "ec2:AllocateAddress",
                "ec2:ReleaseAddress",
                "ec2:DescribeIamInstanceProfileAssociations",
                "ec2:DescribeInstanceStatus",
                "ec2:DescribePlacementGroups",
                "ec2:DescribePrefixLists",
                "ec2:DescribeReservedInstancesOfferings",
                "ec2:DescribeSpotInstanceRequests",
                "ec2:DescribeSpotPriceHistory",
                "elasticfilesystem:DescribeMountTargets",
            },
        ),
        pytest.param(DEFAULT_RAY_IAM_ASSUME_ROLE_POLICY, {"sts:AssumeRole"},),
        pytest.param(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": ["ec2.amazonaws.com"]},
                        "Action": ["sts:AssumeRole", "sts:AssumeRole"],
                    },
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": ["ec2.amazonaws.com"]},
                        "Action": ["sts:AssumeRole", "sts:AssumeRole"],
                    },
                ],
            },
            {"sts:AssumeRole"},
        ),
        pytest.param(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Deny",
                        "Principal": {"Service": ["ec2.amazonaws.com"]},
                        "Action": ["sts:AssumeRole", "sts:AssumeRole"],
                    },
                    {
                        "Effect": "Deny",
                        "Principal": {"Service": ["ec2.amazonaws.com"]},
                        "Action": ["sts:AssumeRole", "sts:AssumeRole"],
                    },
                ],
            },
            set(),
        ),
    ],
)
def test_filter_actions_from_policy_document(policy_document, expected_actions):
    actions = filter_actions_from_policy_document(policy_document=policy_document)
    assert actions == expected_actions


@pytest.mark.parametrize(
    ("assume_role_policy_document", "expected_result"),
    [
        pytest.param(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"AWS": "000000"},
                        "Action": "sts:AssumeRole",
                    },
                ],
            },
            True,
        ),
        pytest.param(
            get_anyscale_aws_iam_assume_role_policy(anyscale_aws_account="000000"),
            True,
        ),
        pytest.param(
            {
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {
                            "AWS": [
                                "arn:aws:iam::other_account",
                                "arn:aws:iam::000000:root",
                            ]
                        },
                        "Action": "sts:AssumeRole",
                    },
                ],
            },
            True,
        ),
        pytest.param(
            {
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {
                            "AWS": [
                                "arn:aws:iam::other_account",
                                "arn:aws:iam::another_account:root",
                            ]
                        },
                        "Action": "sts:AssumeRole",
                    },
                ],
            },
            False,
        ),
        pytest.param(
            {
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "arn:aws:iam::000000:root"},
                        "Action": "sts:AssumeRole",
                    },
                ],
            },
            False,
        ),
    ],
)
def test_contains_control_plane_role(assume_role_policy_document, expected_result):
    result = contains_control_plane_role(
        assume_role_policy_document=assume_role_policy_document,
        anyscale_aws_account="000000",
    )
    assert result == expected_result


@pytest.mark.parametrize(
    ("assume_role_policy_document", "expected_result"),
    [
        pytest.param(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"AWS": "000000"},
                        "Action": "sts:AssumeRole",
                    },
                ],
            },
            False,
        ),
        pytest.param(
            {
                "Statement": [
                    {
                        "Effect": "Reject",
                        "Principal": {
                            "AWS": [
                                "arn:aws:iam::other_account",
                                "arn:aws:iam::000000:root",
                            ]
                        },
                        "Action": "sts:AssumeRole",
                    },
                ],
            },
            False,
        ),
        pytest.param(
            {
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {
                            "AWS": [
                                "arn:aws:iam::other_account",
                                "arn:aws:iam::another_account:root",
                            ]
                        },
                        "Action": "sts:AssumeRole",
                    },
                ],
            },
            False,
        ),
        pytest.param(
            {
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "ec2.amazonaws.com", "AWS": "000000"},
                        "Action": "sts:AssumeRole",
                    },
                ],
            },
            True,
        ),
    ],
)
def test_verify_data_plane_role_assume_role_policy(
    assume_role_policy_document, expected_result
):
    result = verify_data_plane_role_assume_role_policy(
        assume_role_policy_document=assume_role_policy_document
    )
    assert result == expected_result


def test_gen_valid_name():
    name_without_prefiex = gen_valid_name()
    name_with_prefiex = gen_valid_name("test prefix name")

    assert re.match(_NAME_VALIDATION_REGEX_PATTERN, name_without_prefiex) is not None
    assert re.match(_NAME_VALIDATION_REGEX_PATTERN, name_with_prefiex) is not None


@pytest.mark.parametrize(
    ("describe_cluster_response", "raise_exception", "exception_substr"),
    [
        pytest.param({}, True, "does not exist"),
        pytest.param({"Clusters": [{}]}, True, "does not exist"),
        pytest.param(
            {
                "Clusters": [
                    {
                        "Status": "creating",
                        "ARN": "mock_cluster_arn",
                        "ClusterEndpoint": {
                            "Address": "mock_cluster_endpoint_address",
                            "Port": 6379,
                        },
                    }
                ]
            },
            True,
            "not currently available.",
        ),
        pytest.param(
            {
                "Clusters": [
                    {
                        "Status": "available",
                        "ARN": "mock_cluster_arn",
                        "ClusterEndpoint": {
                            "Address": "mock_cluster_endpoint_address",
                            "Port": 6379,
                        },
                    }
                ]
            },
            False,
            "",
        ),
    ],
)
def test_get_memorydb_cluster_config(
    describe_cluster_response, raise_exception, exception_substr
):

    mock_memorydb_cluster_id = "mock_memorydb_cluster_id"
    mock_region = "mock_region"
    mock_memorydb_client = Mock(
        describe_clusters=Mock(return_value=describe_cluster_response)
    )
    mock_boto3 = Mock(client=Mock(return_value=mock_memorydb_client))

    with patch("anyscale.util.boto3", new=mock_boto3):
        if raise_exception:
            with pytest.raises(ClickException) as exc_info:
                _get_memorydb_cluster_config(
                    mock_memorydb_cluster_id, mock_region, Mock()
                )
            assert exception_substr in str(exc_info.value)
        else:
            cluster_config = _get_memorydb_cluster_config(
                mock_memorydb_cluster_id, mock_region, Mock()
            )
            assert cluster_config.id == "mock_cluster_arn"
            assert (
                cluster_config.endpoint == "rediss://mock_cluster_endpoint_address:6379"
            )


@pytest.mark.parametrize(
    ("cloud_region", "subnet_region", "success"),
    [
        pytest.param("us-east-1", "us-east-1", True),
        pytest.param("us-east-1", "us-east-2", False),
    ],
)
@mock_ec2
def test_get_subnet(cloud_region, subnet_region, success):
    ec2 = boto3.client("ec2", region_name=subnet_region)
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet = ec2.create_subnet(
        VpcId=vpc["Vpc"]["VpcId"],
        CidrBlock="10.0.0.0/24",
        AvailabilityZone=f"{subnet_region}a",
    )
    subnet_arn = subnet["Subnet"]["SubnetArn"]

    if not success:
        mock_subnet = Mock(
            return_value=Mock(
                load=Mock(
                    side_effect=ClientError(
                        {
                            "Error": {
                                "Code": "InvalidSubnetID.NotFound",
                                "Message": "Not Found",
                            }
                        },
                        "describe_subnets",
                    )
                )
            )
        )

    else:
        mock_subnet = Mock()

    mock_ec2_resource = Mock(Subnet=mock_subnet)

    mock_logger = Mock()

    with patch("anyscale.util._resource", new=Mock(return_value=mock_ec2_resource)):
        if success:
            _get_subnet(subnet_arn, cloud_region, mock_logger)
        else:
            with pytest.raises(ClickException) as exc_info:
                _get_subnet(subnet_arn, cloud_region, mock_logger)
            assert "does not exist. Please make sur" in str(exc_info.value)
            mock_logger.log_resource_error.assert_called_with(
                CloudAnalyticsEventCloudResource.AWS_SUBNET,
                CloudSetupError.RESOURCE_NOT_FOUND,
            )

    mock_subnet.assert_called_once_with(subnet_arn)



from typing import Optional

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale._private.sdk.timer import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.llm.models import LLMModelsSDK


class LLMSDK(BaseSDK):
    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
        timer: Optional[Timer] = None,
    ):
        self._models = LLMModelsSDK(logger=logger, client=client, timer=timer)

    @property
    def models(self):
        return self._models



import re


def validate_memorystore_instance_name(memorystore_instance_name: str) -> bool:
    if (
        re.search("projects/.+/locations/.+/instances/.+", memorystore_instance_name)
        is None
    ):
        return False
    return True



from dataclasses import dataclass
import os
import re
from typing import Optional

import pytest

from anyscale.job.models import (
    JobConfig,
    JobQueueConfig,
    JobQueueExecutionMode,
    JobQueueSpec,
    JobRunState,
    JobRunStatus,
    JobState,
    JobStatus,
)


@dataclass
class JobConfigFile:
    name: str
    expected_config: Optional[JobConfig] = None
    expected_error: Optional[str] = None

    def get_path(self) -> str:
        return os.path.join(
            os.path.dirname(__file__), "test_files/job_config_files", self.name
        )


TEST_CONFIG_FILES = [
    JobConfigFile(
        "minimal.yaml", expected_config=JobConfig(entrypoint="python test.py"),
    ),
    JobConfigFile(
        "full.yaml",
        expected_config=JobConfig(
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="test-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            entrypoint="python test.py",
            max_retries=5,
        ),
    ),
    JobConfigFile(
        "points_to_requirements_file.yaml",
        expected_config=JobConfig(
            entrypoint="python test.py", requirements="some_requirements_file.txt",
        ),
    ),
    JobConfigFile(
        "unrecognized_option.yaml",
        expected_error=re.escape(
            "__init__() got an unexpected keyword argument 'bad_option'"
        ),
    ),
    JobConfigFile(
        "new_job_queue.yaml",
        expected_config=JobConfig(
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="test-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            entrypoint="python test.py",
            max_retries=5,
            job_queue_config=JobQueueConfig(
                priority=100,
                job_queue_spec=JobQueueSpec(
                    execution_mode=JobQueueExecutionMode.PRIORITY,  # type: ignore
                    name="test-job-queue",
                    compute_config="my-compute-config:1",
                    idle_timeout_s=3600,
                ),
            ),
        ),
    ),
    JobConfigFile(
        "existing_job_queue.yaml",
        expected_config=JobConfig(
            name="test-name-from-file",
            image_uri="docker.io/library/test:latest",
            compute_config="test-compute-config",
            working_dir="test-working-dir",
            excludes=["test"],
            requirements=["pip-install-test"],
            entrypoint="python test.py",
            max_retries=5,
            job_queue_config=JobQueueConfig(
                priority=1000, target_job_queue_name="test-job-queue",
            ),
        ),
    ),
]


class TestJobConfig:
    def test_invalid_entrypoint(self):
        with pytest.raises(ValueError, match="'entrypoint' cannot be empty."):
            JobConfig()

        with pytest.raises(ValueError, match="'entrypoint' cannot be empty."):
            JobConfig(entrypoint="")

        with pytest.raises(TypeError, match="'entrypoint' must be a string."):
            JobConfig(entrypoint=b"oops")

    def test_invalid_max_retries(self):
        with pytest.raises(TypeError, match="'max_retries' must be an int."):
            JobConfig(entrypoint="python test.py", max_retries="1")

        with pytest.raises(ValueError, match="'max_retries' must be >= 0."):
            JobConfig(entrypoint="python test.py", max_retries=-1)

    def test_options(self):
        config = JobConfig(entrypoint="python test.py")

        options = {
            "name": "test-name",
            "image_uri": "docker.io/libaray/test-image:latest",
            "compute_config": "test-compute-config",
            "requirements": ["pip-install-test"],
            "working_dir": ".",
            "excludes": ["some-path"],
            "max_retries": 100,
        }

        # Test setting fields one at a time.
        for option, val in options.items():
            assert config.options(**{option: val}) == JobConfig(
                entrypoint="python test.py", **{option: val}
            )

        # Test setting fields all at once.
        assert config.options(**options) == JobConfig(
            entrypoint="python test.py", **options
        )

    @pytest.mark.parametrize("config_file", TEST_CONFIG_FILES)
    def test_from_config_file(self, config_file: JobConfigFile):
        if config_file.expected_error is not None:
            with pytest.raises(Exception, match=config_file.expected_error):
                JobConfig.from_yaml(config_file.get_path())

            return

        assert config_file.expected_config == JobConfig.from_yaml(
            config_file.get_path()
        )


class TestJobStatus:
    @pytest.mark.parametrize(
        "state",
        [
            JobState.STARTING,
            JobState.RUNNING,
            JobState.FAILED,
            JobState.SUCCEEDED,
            JobState.UNKNOWN,
        ],
    )  # type: ignore
    def test_version_states(self, state: JobState):
        # id, name, state, config
        assert (
            JobStatus(
                id="test-job-id",
                name="test-job-name",
                state=state,
                config=JobConfig(entrypoint="python test.py"),
                runs=[],
            ).state
            == state
        )

    def test_unknown_states(self):
        with pytest.raises(
            ValueError, match="'NOT_REAL_STATE' is not a valid JobState"
        ):
            JobStatus(
                id="test-job-id",
                name="test-job-name",
                state="NOT_REAL_STATE",
                config=JobConfig(entrypoint="python test.py"),
                runs=[],
            )

    def test_job_runs(self):
        status = JobStatus(
            id="test-job-id",
            name="test-job-name",
            state=JobState.SUCCEEDED,
            config=JobConfig(entrypoint="python test.py"),
            runs=[
                JobRunStatus(name="failed-job-run", state=JobRunState.FAILED),
                JobRunStatus(name="succeeded-job-run", state=JobRunState.SUCCEEDED),
            ],
        )

        run_states = [run.state for run in status.runs]
        assert JobRunState.FAILED in run_states
        assert JobRunState.SUCCEEDED in run_states


class TestJobRunStatus:
    @pytest.mark.parametrize(
        "state",
        [
            JobRunState.STARTING,
            JobRunState.RUNNING,
            JobRunState.FAILED,
            JobRunState.SUCCEEDED,
            JobRunState.UNKNOWN,
        ],
    )  # type: ignore
    def test_version_states(self, state: JobRunState):
        assert JobRunStatus(name="test-job-run-name", state=state).state == state

    def test_unknown_states(self):
        with pytest.raises(
            ValueError, match="'NOT_REAL_STATE' is not a valid JobRunState"
        ):
            JobRunStatus(name="test-job-run-name", state="NOT_REAL_STATE")



import logging

from typing_extensions import Literal

from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from anyscale.util import get_cluster_model_for_current_workspace


logger = logging.getLogger(__name__)
from anyscale.anyscale_pydantic import BaseModel


WORKSPACE_NOTIFICATION_ADDRESS = "http://localhost:8266/simplefileserver/notifications"


class WorkspaceNotificationAction(BaseModel):
    type: Literal["navigate-service", "navigate-workspace-tab", "navigate-external-url"]
    title: str
    value: str


class WorkspaceNotification(BaseModel):
    body: str
    action: WorkspaceNotificationAction


def send_workspace_notification(
    anyscale_api_client: BaseApi, notification: WorkspaceNotification
):

    try:
        if get_cluster_model_for_current_workspace(anyscale_api_client) is None:
            # We are not in a workspace, so we don't want to send a notification
            return
        import requests

        requests.post(WORKSPACE_NOTIFICATION_ADDRESS, json=notification.dict())
    except Exception:
        logger.exception("Failed to send notification to UI")



import subprocess
from unittest.mock import MagicMock, Mock, patch

import click
import pytest

from anyscale.client.openapi_client.models.create_experimental_workspace import (
    CreateExperimentalWorkspace,
)
from anyscale.controllers.workspace_controller import WorkspaceController


@pytest.mark.parametrize("project_id", [None, "fake-project-id"])
@pytest.mark.parametrize(
    "compute_config_cloud_id", ["fake-cloud-id", "fake-another-cloud-id"]
)
@patch(
    "anyscale.controllers.workspace_controller.get_default_project",
    Mock(return_value=Mock(id="fake-default-project-id")),
)
def test_create_workspace(
    mock_auth_api_client, project_id, compute_config_cloud_id
) -> None:

    workspace_controller = WorkspaceController()
    mock_create_workspace_api_v2_experimental_workspaces_post = Mock()
    workspace_controller.api_client.create_workspace_api_v2_experimental_workspaces_post = (
        mock_create_workspace_api_v2_experimental_workspaces_post
    )
    mock_get_compute_template_api_v2_compute_templates_template_id_get = Mock(
        return_value=Mock(result=Mock(config=Mock(cloud_id=compute_config_cloud_id)))
    )
    workspace_controller.api_client.get_compute_template_api_v2_compute_templates_template_id_get = (
        mock_get_compute_template_api_v2_compute_templates_template_id_get
    )

    if compute_config_cloud_id != "fake-cloud-id":
        with pytest.raises(click.ClickException):
            workspace_controller.create(
                name="fake-workspace-name",
                cloud_id="fake-cloud-id",
                compute_config_id="fake-compute-config-id",
                cluster_environment_build_id="fake-cluster-environment-build-id",
                project_id=project_id,
            )
            mock_create_workspace_api_v2_experimental_workspaces_post.assert_not_called()
    else:
        workspace_controller.create(
            name="fake-workspace-name",
            cloud_id="fake-cloud-id",
            compute_config_id="fake-compute-config-id",
            cluster_environment_build_id="fake-cluster-environment-build-id",
            project_id=project_id,
        )
        expected_project_id = (
            "fake-project-id" if project_id else "fake-default-project-id"
        )

        mock_create_workspace_api_v2_experimental_workspaces_post.assert_called_once_with(
            CreateExperimentalWorkspace(
                name="fake-workspace-name",
                cloud_id="fake-cloud-id",
                compute_config_id="fake-compute-config-id",
                cluster_environment_build_id="fake-cluster-environment-build-id",
                project_id=expected_project_id,
            )
        )


@pytest.mark.parametrize("delete_existing_files_in_destination", [True, False])
def test_run_rsync(mock_auth_api_client, delete_existing_files_in_destination: bool):
    # Mock subprocess.run to prevent actual execution of the rsync command
    with patch.object(subprocess, "run") as mocked_run, patch.object(
        WorkspaceController, "_get_workspace_command_context"
    ) as mock_get_workspace_command_context:
        mock_get_workspace_command_context.return_value = MagicMock(
            ssh=MagicMock(key_path="mock_key_path", user="mock_user"),
            head_ip="mock_head_ip",
            working_dir="mock_working_dir",
        )
        workspace_controller = WorkspaceController()

        # Set up the expected rsync command
        expected_rsync_command = [
            "rsync",
            "--rsh",
            'ssh -i mock_key_path -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o IdentitiesOnly=yes -o "ProxyCommand=ssh -W %h:%p -i mock_key_path -o StrictHostKeyChecking=accept-new mock_user@mock_head_ip" -p 5020',
            "-rvzl",
            "--delete",
            "--exclude",
            "exclude_pattern1",
            "--exclude",
            "exclude_pattern2",
            "--filter",
            "dir-merge,- filter_pattern1",
            "--filter",
            "dir-merge,- filter_pattern2",
            "ray@0.0.0.0:mock_remote_path",
            "mock_local_path",
        ]

        if not delete_existing_files_in_destination:
            expected_rsync_command.remove("--delete")

        # Call the run_rsync method with mock parameters
        workspace_controller.run_rsync(
            local_path="mock_local_path",
            remote_path="mock_remote_path",
            down=True,
            rsync_filters=["filter_pattern1", "filter_pattern2"],
            rsync_excludes=["exclude_pattern1", "exclude_pattern2"],
            delete_existing_files_in_destination=delete_existing_files_in_destination,
        )

        # Check if subprocess.run was called with the expected rsync command
        mocked_run.assert_called_once_with(expected_rsync_command, check=False)



import os
import subprocess
from typing import Dict, Generator, Optional, Tuple, Union
import uuid

import click
from click.testing import CliRunner
import pytest

from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.commands.workspace_commands_v2 import (
    create,
    run_command,
    ssh,
    start,
    status,
    terminate,
    wait,
)
from anyscale.workspace.commands import _WORKSPACE_SDK_SINGLETON_KEY
from anyscale.workspace.models import WorkspaceConfig, WorkspaceState


def _get_test_file_path(subpath: str) -> str:
    return os.path.join(os.path.dirname(__file__), "test_files/", subpath)


EMPTY_CONFIG_PATH = _get_test_file_path("workspace_config_files/empty.yaml")
MINIMAL_CONFIG_PATH = _get_test_file_path("workspace_config_files/minimal.yaml")
FULL_CONFIG_PATH = _get_test_file_path("workspace_config_files/full.yaml")
UNRECOGNIZED_OPTION_CONFIG_PATH = _get_test_file_path(
    "workspace_config_files/unrecognized_option.yaml"
)

FULL_CONFIG_SCHEDULE_NAME = "test-name-from-file"


class FakeWorkspaceSDK:
    def __init__(self):
        self._created_workspaces: Dict[str, WorkspaceConfig] = {}
        self._workspace_status = {}

    def _resolve_to_workspace_model(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> Tuple[str, WorkspaceConfig]:
        workspace_config = None
        workspace_id = None
        if id is not None:
            workspace_config = self._created_workspaces.get(id)
            workspace_id = id
        elif name is not None:
            for key, config in self._created_workspaces.items():
                if (
                    config.name == name
                    and (cloud is None or config.cloud == cloud)
                    and (project is None or config.project == project)
                ):
                    workspace_config = config
                    workspace_id = key
                    break
        assert workspace_id is not None, "Workspace was not found."
        if workspace_config is None:
            raise RuntimeError("Workspace was not found.")
        return workspace_id, workspace_config

    def create(self, config: WorkspaceConfig) -> str:
        assert isinstance(config, WorkspaceConfig)
        id = str(uuid.uuid4())  # noqa: A001
        self._created_workspaces[id] = config
        self._workspace_status[id] = WorkspaceState.TERMINATED
        return id

    def start(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_id, _ = self._resolve_to_workspace_model(
            name=name, id=id, cloud=cloud, project=project
        )
        self._workspace_status[workspace_id] = WorkspaceState.RUNNING
        return workspace_id

    def terminate(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_id, _ = self._resolve_to_workspace_model(
            name=name, id=id, cloud=cloud, project=project
        )
        self._workspace_status[workspace_id] = WorkspaceState.TERMINATED
        return workspace_id

    def status(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_id, _ = self._resolve_to_workspace_model(
            name=name, id=id, cloud=cloud, project=project
        )
        return self._workspace_status[workspace_id]

    def wait(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        timeout_s: float = 5,
        state: Union[str, WorkspaceState] = WorkspaceState.RUNNING,
    ):
        workspace_id, _ = self._resolve_to_workspace_model(
            name=name, id=id, cloud=cloud, project=project
        )
        target_state = WorkspaceState.validate(state)
        current_status = self._workspace_status[workspace_id]

        if current_status != target_state:
            raise TimeoutError(f"Workspace did not reach state {target_state} in time.")

    def generate_ssh_config_file(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        ssh_config_path: Optional[str] = None,
    ) -> Tuple[str, str]:
        _, model = self._resolve_to_workspace_model(
            name=name, id=id, cloud=cloud, project=project
        )
        assert model.name is not None
        return model.name, f"{ssh_config_path}/ssh_config"

    def run_command(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        command: str,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        workspace_id, _ = self._resolve_to_workspace_model(
            name=name, id=id, cloud=cloud, project=project
        )
        print(f"Running command {command} for {workspace_id}")
        return subprocess.CompletedProcess(
            args=[],
            returncode=0,
            stdout=f"Output for {workspace_id} and {command}".encode("utf-8"),
            stderr=b"",
        )

    def get_default_dir_name(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_id, _ = self._resolve_to_workspace_model(
            name=name, id=id, cloud=cloud, project=project
        )
        return "/home/ray/default"


@pytest.fixture()
def fake_workspace_sdk() -> Generator[FakeWorkspaceSDK, None, None]:
    fake_workspace_sdk = FakeWorkspaceSDK()
    _LAZY_SDK_SINGLETONS[_WORKSPACE_SDK_SINGLETON_KEY] = fake_workspace_sdk
    try:
        yield fake_workspace_sdk
    finally:
        del _LAZY_SDK_SINGLETONS[_WORKSPACE_SDK_SINGLETON_KEY]


def _assert_error_message(result: click.testing.Result, *, message: str):
    assert result.exit_code != 0
    assert message in result.stdout


class TestCreate:
    def test_missing_arg(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create)
        _assert_error_message(result, message="Workspace name must be configured")

    def test_config_file_not_found(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", "missing_config.yaml"])
        _assert_error_message(
            result, message="Error: Config file 'missing_config.yaml' not found.",
        )

    @pytest.mark.parametrize(
        "config_file_arg", [MINIMAL_CONFIG_PATH, FULL_CONFIG_PATH],
    )
    def test_basic(self, fake_workspace_sdk, config_file_arg):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", config_file_arg])
        assert result.exit_code == 0, result.stdout
        assert len(fake_workspace_sdk._created_workspaces) == 1

    def test_override_name(self, fake_workspace_sdk):
        runner = CliRunner()
        name = "test-different-name"
        result = runner.invoke(create, ["--name", name, "-f", FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert result.exit_code == 0, result.stdout
        assert len(fake_workspace_sdk._created_workspaces) == 1

        workspace_config = list(fake_workspace_sdk._created_workspaces.values())[0]
        assert workspace_config.name == name


class TestStart:
    def test_missing_arg(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(start)
        _assert_error_message(
            result, message="Error: One of '--name' and '--id' must be provided.\n"
        )

    def test_start_with_id(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        # Start the workspace
        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        result = runner.invoke(start, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout
        assert (
            fake_workspace_sdk._workspace_status[workspace_id] == WorkspaceState.RUNNING
        )

    def test_start_with_name(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        # Start the workspace
        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name
        result = runner.invoke(start, ["--name", workspace_name])
        assert result.exit_code == 0, result.stdout
        assert (
            fake_workspace_sdk._workspace_status[workspace_id] == WorkspaceState.RUNNING
        )

    def test_invalid_args(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name

        result = runner.invoke(start, ["--name", workspace_name, "--id", workspace_id])
        _assert_error_message(
            result, message="Only one of '--name' and '--id' can be provided."
        )


class TestTerminate:
    def test_missing_arg(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(terminate)
        _assert_error_message(
            result, message="Error: One of '--name' and '--id' must be provided.\n"
        )

    def test_terminate_with_id(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        # Terminate the workspace
        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        result = runner.invoke(terminate, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout
        assert (
            fake_workspace_sdk._workspace_status[workspace_id]
            == WorkspaceState.TERMINATED
        )

    def test_terminate_with_name(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        # Terminate the workspace
        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name
        result = runner.invoke(terminate, ["--name", workspace_name])
        assert result.exit_code == 0, result.stdout
        assert (
            fake_workspace_sdk._workspace_status[workspace_id]
            == WorkspaceState.TERMINATED
        )

    def test_invalid_args(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name

        result = runner.invoke(
            terminate, ["--name", workspace_name, "--id", workspace_id]
        )
        _assert_error_message(
            result, message="Only one of '--name' and '--id' can be provided."
        )


class TestStatus:
    def test_missing_arg(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(status)
        _assert_error_message(
            result, message="Error: One of '--name' and '--id' must be provided.\n"
        )

    def test_status_with_id(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        result = runner.invoke(status, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout
        assert (
            fake_workspace_sdk._workspace_status[workspace_id]
            == WorkspaceState.TERMINATED
        )

    def test_status_with_name(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name
        result = runner.invoke(status, ["--name", workspace_name])
        assert result.exit_code == 0, result.stdout
        assert (
            fake_workspace_sdk._workspace_status[workspace_id]
            == WorkspaceState.TERMINATED
        )

    def test_invalid_args(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name

        result = runner.invoke(status, ["--name", workspace_name, "--id", workspace_id])
        _assert_error_message(
            result, message="Only one of '--name' and '--id' can be provided."
        )


class TestWait:
    def test_missing_arg(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(wait)
        _assert_error_message(
            result, message="Error: One of '--name' and '--id' must be provided.\n"
        )

    def test_wait_with_id(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        result = runner.invoke(
            wait, ["--id", workspace_id, "--state", WorkspaceState.TERMINATED]
        )
        assert result.exit_code == 0, result.stdout
        assert (
            fake_workspace_sdk._workspace_status[workspace_id]
            == WorkspaceState.TERMINATED
        )

    def test_wait_with_name(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name
        result = runner.invoke(
            wait, ["--name", workspace_name, "--state", WorkspaceState.TERMINATED]
        )
        assert result.exit_code == 0, result.stdout
        assert (
            fake_workspace_sdk._workspace_status[workspace_id]
            == WorkspaceState.TERMINATED
        )

    def test_invalid_args(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name

        result = runner.invoke(wait, ["--name", workspace_name, "--id", workspace_id])
        _assert_error_message(
            result, message="Only one of '--name' and '--id' can be provided."
        )


class TestSSH:
    def test_missing_arg(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(ssh)
        _assert_error_message(
            result, message="Error: One of '--name' and '--id' must be provided.\n"
        )

    def test_ssh_with_id(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        # start it
        result = runner.invoke(start, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(ssh, ["--id", workspace_id, "--command", "ls"])
        assert result.exit_code == 0, result.stdout

    def test_ssh_with_name(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name

        # start it
        result = runner.invoke(start, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(ssh, ["--name", workspace_name])
        assert result.exit_code == 0, result.stdout

    def test_invalid_args(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name

        result = runner.invoke(ssh, ["--name", workspace_name, "--id", workspace_id])
        _assert_error_message(
            result, message="Only one of '--name' and '--id' can be provided."
        )

    def test_with_ssh_args(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        # start it
        result = runner.invoke(start, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(
            ssh, ["--id", workspace_id, "--", "-l", "9000:localhost:9000"]
        )
        assert result.exit_code == 0, result.stdout


class TestRunCommand:
    def test_missing_arg(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        # start it
        result = runner.invoke(start, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(ssh)
        _assert_error_message(
            result, message="Error: One of '--name' and '--id' must be provided.\n"
        )

    def test_run_command_with_id(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        # start it
        result = runner.invoke(start, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(run_command, ["--id", workspace_id, "ls"])
        assert result.exit_code == 0, result.stdout
        assert "ls" in result.stdout

    def test_run_command_with_name(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name
        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]

        # start it
        result = runner.invoke(start, ["--id", workspace_id])
        assert result.exit_code == 0, result.stdout

        result = runner.invoke(run_command, ["--name", workspace_name, "ls"])
        assert result.exit_code == 0, result.stdout
        assert "ls" in result.stdout

    def test_invalid_args(self, fake_workspace_sdk):
        runner = CliRunner()
        result = runner.invoke(create, ["-f", MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout

        workspace_id = list(fake_workspace_sdk._created_workspaces.keys())[0]
        workspace_name = list(fake_workspace_sdk._created_workspaces.values())[0].name

        result = runner.invoke(
            run_command, ["--name", workspace_name, "--id", workspace_id, "ls"]
        )
        _assert_error_message(
            result, message="Only one of '--name' and '--id' can be provided."
        )
